/* storage4.c */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/

#include "global.h" 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
010215 storage4.c update MixPlug for hydraulic inlets, no external pipe2 any more
       MixedPlug Model optional for HE1 and HE2 (heatexchager)
010221 last_time_flag losses print
010222 convection updated,Qi[i][3] bug fix
010306 initial value pipe_HE2_exist=0 set (bug)
010523 if mode==3 pipe1_exist set,energy history check increased to 2Ws
010625 history_check:Qinit[storage_counter] ... -Qunit_transfer;

*/

int internal_pipe(double *dQout,double *dQloss,double dQin,double mp_cp_h,
                  int current_node,int i_offset,double Tamb,double kA,
                  double mc_i,double ***qp_sum,int unit,
                  double energy[STORAGE_MAX][DERIV_MAX],int index,
                  double **deriv_o,double **deriv_n,double **init_energy,
                  double h);
int GetStratisfiedInletNode(double dQin,double mp_cp_h,int n,int unit,double **deriv_o);
int MixPlug(double *dQout,double dQin,double mp_cp_h,double mc,int current_node,
            int in_node,int out_node,double **deriv_o,double **deriv_n,int unit);
int GetStratisfiedInletNode(double dQin,double mp_cp_h,int n,int unit,double **deriv_o);

/*-----------Storage4 -------------------------------------------------------*/
int storage (unit,par,in_u,in_nr,out_n,out_o,deriv_o,deriv_n,h,
             init_energy,qp_sum,tsim,init_flag,type,day_of_year,
             previous_unit,un_max,select_unit,type_index)
           int    unit;      /* current unit number */
           double **par;     /* parameter array par[unit][1..PARMAX]  */
           int    **in_u;    /* input unit array, input connections   */
           int    **in_nr;   /* input number array                    */
           double **out_o;   /* output array with privious state      */
           double **out_n;   /* output array with current state       */
           double **deriv_o; /* derivatives: used for diffential equ. */
           double **deriv_n; /*  new state of deriv                   */
           double h;         /* sim time step                         */
           double **init_energy; /* initial energy in RC-comp. [Ws]   */
           double ***qp_sum; /* energy to one node qp[unit][n][m]     */
           double tsim;      /* simulation time in [s] of this day    */
           int    init_flag; /* set at initial loop of sim            */
           int    *type;     /* type number array                     */
           int    day_of_year;
           int    previous_unit;
           int    un_max;    /* unit max */
           int    *select_unit; /* number of calls in this timestep   */
           int    *type_index; /* index of calling this type          */
{
/* xfig.awk

   TYPE 04
   PRIORITY  6          used by config routine
   HYDRAULIC 4          4-pole fluid            


   PAR1: 10 [W/K] kA ambient losses
   PAR2: 1000 [l] volume of storage incl heatexchangers
   PAR3: 4.19 [kJ/kgK] cp of storage fluid 
   PAR4: 10 [] number n of nodes
   PAR5: 3 [] inlet  node of heater cycle (HE1)
   PAR6: 5 [] outlet node of heater cycle (HE1)
   PAR7: 4.19 [kJ/kgK] cp of fluid from heater cycle
   PAR8: 4 [m] l1 length of HE1 (heater cycle)
   PAR9: 15 [mm] d1 diameter HE1 [mm]
   PAR10: 50 [W/Km] kA1/l linear heat transfer coeff HE1, no fit
   PAR11: 6 [] inlet  node of solar cycle (HE2) 
   PAR12: 9 [] outlet node of solar cycle (HE2) 
   PAR13: 4.19 [W/kgK] cp of fluid from solar cycle (HE2)
   PAR14: 4 [m] l2 length of HE2 (solar cycle)
   PAR15: 15 [mm] d2 diameter HE2 [mm]
   PAR16: 50 [W/Km] kA2/l,<0:internal FIT with T_HE2_amb HE2_out,with scaling
   PAR17: 1 [] node of temp. sensor 1
   PAR18: 2 [] node of temp. sensor 2
   PAR19: 8 [] node of temp. sensor 3
   PAR20: 10 [] node of temp. sensor 4
   PAR21: 3 [] circulation node inlet
   PAR22: 0 [] mode0: 2 int.HE mode1:0 int.HE mode2:1 int.HE (solar)
   PAR23: 1 [] outlet node of warm water, if 0-->par23=1
   PAR24: 10 [] inlet node of fresh water,if 0 then par24=par4
   PAR25: 4000 [W/K] kA_convection.If 0: Maximal mixing
   PAR26: 100 [W/K]  kA_heat_conduction.If 0:Calculated from lambda=0.25
   PAR27: 2.5 [m] storage height, used to modify ambient losses(circ.tanks)
   PAR28: 1 [1] 0:no stratification 1:with stratification
   
   INP1: 0 [Ws] heat flow inlet solar cycle (HE2)
   INP2: 0 [kg/s] mass flow inlet solar cycle (HE2)
   INP3: 0 [Ws] heat flow inlet fresh water 
   INP4: 0 [kg/s] mass flow inlet fresh water
   INP5: 0 [Ws] heat flow inlet heater cycle (HE1)
   INP6: 0 [kg/s] mass flow inlet heater cycle (HE2)
   INP7: 0 [Ws] heat flow inlet circulation
   INP8: 0 [kg/s] mass flow inlet circulation
   INP9: 20.0 [C] ambient temperature

   OUT1: [Ws] 'dQHE2' heat flow outlet solar cycle (HE1)
   OUT2: [kg/s] 'mpHE2' mass flow outlet solar cycle (HE1)
   OUT3: [Ws] 'dQww' heat flow outlet warm water
   OUT4: [kg/s] 'mp_ww' mass flow outlet warm water 
   OUT6: [Ws] 'sys_gain' system gains 
   OUT7: [Ws] 'sys_loss' system losses
   OUT8: [Ws] 'sys_energy' internal energy
   
   OUT9: [Ws]  'dQHE1' heat flow outlet heater cycle (HE1)
   OUT10: [kg/s] 'mpHE1' mass flow outlet heater cycle (HE1)
   OUT11: [C] 'T_sensor1'    node par17
   OUT12: [C] 'T_sensor2'    node par18
   OUT13: [C] 'T_sensor3'    node par19
   OUT14: [C] 'T_sensor4'    node par20
   OUT17: [Ws] 'dQcirc' heat flow outlet circulation
   OUT18: [kg/s] 'mpcirc' mass flow outlet circulation
   OUT19: [] -   
   OUT20: [W] 'PHE2' Power heatexch. HE2
   OUT21: [W] 'PHE1' Power heatexch. HE1
   OUT22: [W] 'Pww' Power of load warm water
   OUT23: [W] 'Ploss' losses to ambient
   OUT24: [C] 'ToutHE2' temp HE2 out 
   OUT25: [C] 'Tmean' T_mean 1..n                      
   OUT26: [Ws] Q_lower_part HE2 and node 6..10
   OUT27: [W] 'Pcirc' Power of circulation
   OUT28: [C] 'TinHE2' temp HE2 in
   OUT29: [C] 'TinHE1' temp HE1 inlet
   OUT30: [C] 'ToutHE1' temp HE1 outlet
 
   DERIV1: 60 [C] 'T1' temperatur of node 1 (top) init_energy 1=DERIV1/mc
   DERIV2: 60 [] 'T2'
   DERIV3: 60 [] 'T3'
   DERIV4: 60 [] 'T4'
   DERIV5: 60 [] 'T5'  
   DERIV6: 60 [] 'T6'  
   DERIV7: 60 [] 'T7'  
   DERIV8: 60 [] 'T8'  
   DERIV9: 60 [] 'T9'  
   DERIV10: 60 [] 'T10'

*/

int a=0; /*-------------ANALYSE a=1 ---------*/
int pedantic=1;
static int init_init=1; /* initialize int init_flag_local[] array */
static int init_flag_local[STORAGE_MAX];
static int store_call_nr[STORAGE_MAX]; /* inc that by init process */
static double energy[STORAGE_MAX][DERIV_MAX];
int last_time_flag; /* check number of calls vai select */
int storage_counter=type_index[unit]; /* cmp wal__index in wall.c */

int               n=(int) par[unit][4];       /* number of nodes */
double            kA_node=par[unit][1]/ ( (double) n) ;
double                Vol=par[unit][2];
double                 cp=par[unit][3]*1e+3;
int         HE1_in =(int) par[unit][5];  /* HE1_in node of inlet HE1 */
int         HE1_out=(int) par[unit][6];  /* oulet-node of HE1 */
double                cp1=par[unit][7]*1e+3;

double                 d1=par[unit][9]; /*diameter HE1 */
double                 l1=par[unit][8]; /*length of HE1 */
double           kA_l_HE1=par[unit][10]; 

int         HE2_in =(int) par[unit][11];  /* inlet-node of HE2 */
int         HE2_out=(int) par[unit][12]; /* oulet-node of HE2 */
double                cp2=par[unit][13]*1e+3;
double                 l2=par[unit][14]; /*length of HE2 */
double                 d2=par[unit][15]; /*diameter HE2 */
double           kA_l_HE2=par[unit][16]; 
int    node_sensor1=(int) par[unit][17];
int    node_sensor2=(int) par[unit][18];
int    node_sensor3=(int) par[unit][19];
int    node_sensor4=(int) par[unit][20];
int    circ_node_in=(int) par[unit][21];

int            mode=(int) par[unit][22]; /* 0 [] mode 0:2WT 1:0WT 2:1WT (solar) */
int     cw_node_out=(int) par[unit][23]; /* 0 [] mode 0:2WT 1:0WT 2:1WT (solar) */
int      cw_node_in=(int) par[unit][24]; /* cw_node_in */
double        kA_down_up= par[unit][25];
double kA_heat_conduction=par[unit][26];
double      StorageHeight=par[unit][27];
int        stratify=(int) par[unit][28];

double   dQ_HE2   =out_n IN (1);
double     mp_HE2 =out_n IN (2);
double mp_cp_HE2_h=out_n IN (2)*cp2*h;
double dQ_cw_in   =out_n IN (3);
double mp_cp_cw_h =out_n IN (4)*cp*h;
double     dQ_HE1 =out_n IN (5);
double     mp_HE1  =out_n IN (6);
double mp_cp_HE1_h =out_n IN (6)*cp1*h;
double dQ_circ_in  =out_n IN (7);
double mp_cp_circ_h=out_n IN (8)*cp*h;
double Tamb        =out_n IN (9);

int    circ_node_out=1;
int dir;
double Qi[NODE_MAX][10];
double kA_node_corr=kA_node; /* used if StorageHeight>0 */
double mc_node; /* vol*density*cp of storage water  */
double Q_help;
int i_node,i_pipe_HE1_offset,i_pipe_HE2_offset;
int pipe_HE1_exist=0,pipe_HE2_exist=0;

double Vol1,Vol2,kA_HE1,kA_HE2;
double mc_pipe1,mc_pipe2;
int max_circ,min_circ;
int max_cw,min_cw;
int min_HE1,max_HE1,min_HE2,max_HE2,i,i1,n_pipe1=0,n_pipe2=0;
int call_HE1_flag,call_HE2_flag,circ_flag,KW_flag;
int first_time_flag=0; /* first call of unit in timestep */
double Qmp_in;
double Qout_HE1=0,Qout_HE2=0;
double P_WW; /* ,PHE1,PHE2,Ploss; */
double TinHE1,TinHE2,ToutHE1,ToutHE2;



if (init_init) { /* only once for all storages */
 init_init=0; for (i=0;i<STORAGE_MAX;i++) init_flag_local[i]=1;
}

 /* mode 0:2WT 1:0WT 2:1WT (solar) */
if ( (mode==2) || (mode==0) ) pipe_HE2_exist=1;
if ( (mode==0) || (mode==3) ) pipe_HE1_exist=1;

/* 
pipe_HE1_exist=1; 
pipe_HE2_exist=1; 
 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!11 */

if (a) printf("storage\n");
if (a) fflush(stdout);

/* check first time call in this timestep */
if (select_unit[unit]++==0)  first_time_flag=1; /* multiple call type */
if (init_flag) store_call_nr[storage_counter]=select_unit[unit]; /* get select_unit_max */
if (store_call_nr[storage_counter]==select_unit[unit]) last_time_flag=1;
else                                                   last_time_flag=0;

if (a) printf("tsim:%.1f call:%d storage_counter:%d store_call_nr[storage_counter]:%d "
    "last_time_flag:%d\n"
    ,tsim,select_unit[unit],storage_counter,store_call_nr[storage_counter],last_time_flag);
              
if ( (stratify) && (pipe_HE2_exist) ) {
 printf("storage:%d stratify:%d pipe_HE2_exist:%d\n",unit,stratify,pipe_HE2_exist);
 exit (-1);
}
if ( (stratify) && (mp_HE2>0) && (!pipe_HE2_exist) ) {
  HE2_in=GetStratisfiedInletNode(dQ_HE2,mp_cp_HE2_h,n,unit,deriv_o);
  HE2_out=n;
}

  /* select input-unit for storage (storage is a multi-called unit !!) */
call_HE1_flag=( (HE1_in) && (in_u[unit][5]==previous_unit) ) /* call up hydr circle */
           || ( (in_u[unit][5]==unit) && (first_time_flag) );

call_HE2_flag=(    (HE2_in) && (in_u[unit][1]==previous_unit))
           || (    (in_u[unit][1]==unit) && (first_time_flag));
               
circ_flag=    (     in_u[unit][7]==previous_unit)
           || (    (in_u[unit][7]==unit) && (first_time_flag));

KW_flag  =    (     in_u[unit][3]==previous_unit)
           || (    (in_u[unit][3]==unit) && (first_time_flag));

  /* Volume of heat-exchangers WTs in [l] */
Vol1=PI*d1*d1/4*l1*1e-3;
Vol2=PI*d2*d2/4*l2*1e-3;

  /* heat-capacity per node */
mc_node = Vol*cp/( (double) n); /* brutto volume */
if (mode!=1) mc_node-=Vol2*cp/( (double) n); /* netto without hex's */
if (mode==0) mc_node-=Vol1*cp/( (double) n);

if (a) printf("mc_node: %.4f \n",mc_node);

if (circ_node_in>circ_node_out) { max_circ=circ_node_in; min_circ=circ_node_out; }
else                            { max_circ=circ_node_out;min_circ=circ_node_in;  }
if (cw_node_in>cw_node_out)     { max_cw=cw_node_in; min_cw=cw_node_out; }
else                            { max_cw=cw_node_out;min_cw=cw_node_in;  }

if (HE1_out>=HE1_in) { min_HE1=HE1_in; max_HE1=HE1_out; }
else                 { min_HE1=HE1_out;max_HE1=HE1_in;  }
if (HE2_out>=HE2_in) { min_HE2=HE2_in; max_HE2=HE2_out; }
else                 { min_HE2=HE2_out;max_HE2=HE2_in;  }

if (pipe_HE1_exist) { 
  n_pipe1=max_HE1-min_HE1+1;
  mc_pipe1=Vol1*cp1/( (double) n_pipe1);
  kA_HE1=kA_l_HE1*l1/( (double) n_pipe1); /* kA-total */
  /* i_pipe_HE1_offset=min_HE1-(n+1); */
  i_pipe_HE1_offset=n-min_HE1+1; /*i_node=i_current+i_pipe_HE1_offset */
}
if (pipe_HE2_exist)  {
  double Tpipe,Tpipe_ambient;
  n_pipe2=max_HE2-min_HE2+1;
  mc_pipe2=Vol2*cp2/( (double) n_pipe2);
  kA_HE2=kA_l_HE2*l2/( (double) n_pipe2); /* per adjacent node ! */
  i_pipe_HE2_offset=n-min_HE2+1+n_pipe1; /* i_node=i_current+offset:  */
  Tpipe=deriv_o[unit][min_HE2];
  Tpipe_ambient=deriv_o[unit][min_HE2+i_pipe_HE2_offset];
  if (kA_l_HE2==-1) /* use ITW fit */
   kA_HE2=9.0*4.0*pow(fabs(Tpipe_ambient-Tpipe),0.6);
  if (min_HE2!=max_HE2) { 
    printf("storage4: min_HE2:%d max_HE2:%d\n",min_HE2,max_HE2); exit (-1); }
  if (kA_HE2>200) printf("kA_HE2:%.2f Tp[%d]:%.2f T[%d]:%.2f kA:%.2f\n"
     ,kA_HE2,min_HE2+i_pipe_HE2_offset,deriv_o[unit]
      [min_HE2+i_pipe_HE2_offset],min_HE2,deriv_o[unit][min_HE2],kA_HE2);
}
if (a) printf("storage\n");

/*------ init-----------------------------------------------------------------*/
if (init_flag_local[storage_counter]) { /* call only once per storage*/
  int i;
  double q_storage=0,T;

  init_flag_local[storage_counter]=0; /* only once */ 

  if ( cw_node_in==0) { printf("par[%d][24] not set %d\n",unit,cw_node_in); }
  if ( cw_node_out==0) { printf("par[%d][23] not set  %d\n",unit,cw_node_out);   } 

      /*------ check WT-parameter */
  if (   (HE1_in >n) || (HE1_in<1) ||(HE1_out>n) || (HE1_out<1)
       ||(HE2_in >n) || (HE2_in <1) ||(HE2_out>n) || (HE2_out<1)
     ) { printf("exit, caused by WT-Parameters\n"); exit(-1); }
  if ( (mode<0) || (mode>3) ) {
    printf("storage:%d error mode:%d\n",unit,mode); exit(-1); }


  if ( deriv_o[unit][1]==0 ) printf(" set deriv1 of storage unit:%d",unit);

      /*------ init energy */
  T=10; /* default temperature */
  for (i=1;i<=n;i++) {
    if ( deriv_n[unit][i]>0) T=deriv_o[unit][i]=deriv_n[unit][i];
    else deriv_o[unit][i]=deriv_n[unit][i]=T; /* at least deriv1!=0 */
    init_energy[unit][i]=T*mc_node;
    q_storage+=init_energy[unit][i];
    if (a) printf("deriv_o: %.4f  deriv_n: %.4f \n" ,deriv_o[unit][i] ,deriv_n[unit][i]);
  };

  if (pipe_HE1_exist) { 
    for (i=min_HE1;i<=max_HE1;i++) {
     i_node=i_pipe_HE1_offset+i; /* adjacent storage node */
     T=deriv_n[unit][i_node-i_pipe_HE1_offset]; 
     deriv_n[unit][i_node]=deriv_o[unit][i_node]= T;
     energy[storage_counter][i_node]=init_energy[unit][i_node]=T*mc_pipe1;
    }
  } /* init pipe2/HE2 */
  if (pipe_HE2_exist) { 
    for (i=min_HE2;i<=max_HE2;i++) {
     i_node=i_pipe_HE2_offset+i; /* adjacent storage node */
     T=deriv_n[unit][i_node-i_pipe_HE2_offset]; /* get node temp in order to copy to pipe_node */
     deriv_n[unit][i_node]=deriv_o[unit][i_node]= T;
     energy[storage_counter][i_node]=init_energy[unit][i_node]=T*mc_pipe2;
    }
  } /* init pipe2/HE2 */
}; /* init */
/*---end init-----------------------------------------------------------------*/
/* if (init_flag) return (1); / * last_time_flag */
if (a) printf("tsim:%.1f storage[%d] 2.storage storage_counter:%d  \n"
              ,tsim,unit,storage_counter);
               fflush(stdout);

/* Qi[i][1]:HE1_inlet;Qi[i][2]:kAdown-up;Qi[i][3]:kAambient;Qi[i][4]:KWin,Qi[i][5]:WWout,
   Qi[i][6]:CircIn; Qi[i][7]: CircOut; Qi[i][8]:HE2, */


if (first_time_flag) {
}
if (call_HE1_flag) { /* calc HE1 */
  if (HE1_in<=HE1_out) dir=1; else dir=-1;
  i=HE1_in;
  do { /* walk thru nodes */
    if (i==HE1_in) Qmp_in=dQ_HE1;
    else           Qmp_in=Qout_HE1;
    if (pipe_HE1_exist) internal_pipe(&Qout_HE1,&Qi[i][1],Qmp_in,mp_cp_HE1_h,i,
                                            i_pipe_HE1_offset,deriv_o[unit][i],
                                            kA_HE1,mc_pipe1,qp_sum,unit,energy,
                                            storage_counter,deriv_o,deriv_n,
                                            init_energy,h);
    else { 
      MixPlug(&Qout_HE1,Qmp_in,mp_cp_HE1_h,mc_node,i,HE1_in,HE1_out,
              deriv_o,deriv_n,unit);
      Qout_HE1*=-1;
      Qi[i][1]=Qmp_in-Qout_HE1;
      }
    qp_sum[unit][i][1]+=Qi[i][1];
    i+=dir;
  } while ( (min_HE1<=i) && (i<=max_HE1) ) ;
  if (mp_cp_HE1_h>0) TinHE1=dQ_HE1/(mp_cp_HE1_h+1e-12);
  else               TinHE1=deriv_o[unit][HE1_in];
  if (mp_cp_HE1_h>0) ToutHE1=Qout_HE1/(mp_cp_HE1_h+1e-12);
  else               ToutHE1=deriv_o[unit][HE1_out];
} /* call_HE1_flag */

if (call_HE2_flag) { /* calc HE1 */
 if (HE2_in<=HE2_out) dir=1; else dir=-1;
 i=HE2_in;
 do { /* walk thru nodes */
   if (i==HE2_in) Qmp_in=dQ_HE2;
   else           Qmp_in=Qout_HE2;
   if (pipe_HE2_exist) internal_pipe(&Qout_HE2,&Qi[i][8],Qmp_in,mp_cp_HE2_h,i,
                                      i_pipe_HE2_offset,deriv_o[unit][i],
                                      kA_HE2,mc_pipe2,qp_sum,unit,energy,
                                      storage_counter,deriv_o,deriv_n,
                                      init_energy,h);
   else {
      MixPlug(&Qout_HE2,Qmp_in,mp_cp_HE2_h,mc_node,i,HE2_in,HE2_out,
          deriv_o,deriv_n,unit);
      Qout_HE2*=-1;
      Qi[i][8]=Qmp_in-Qout_HE2;
   }
   qp_sum[unit][i][8]+=Qi[i][8];
   i+=dir;
   } while ( (min_HE2<=i) && (i<=max_HE2) ) ;
  TinHE2=dQ_HE2/(mp_cp_HE2_h+1e-12);
  ToutHE2=Qout_HE2/(mp_cp_HE2_h+1e-12);
} /* call_HE2_flag */ 

if (circ_flag)  {
  if (circ_node_in<=circ_node_out) dir=1; else dir=-1;
  i=circ_node_in;
  do { /* walk thru nodes */
    if (i==circ_node_in) Qi[i][6]=dQ_circ_in; 
    else                 Qi[i][6]=-Qi[i-dir][7]; 
    MixPlug(&Qi[i][7],Qi[i][6],mp_cp_circ_h,mc_node,i,circ_node_in,circ_node_out,
            deriv_o,deriv_n,unit);
    qp_sum[unit][i][6]+=Qi[i][6];
    qp_sum[unit][i][7]+=Qi[i][7];
    i+=dir; 
  } while ( (min_circ<=i) && (i<=max_circ) ) ;
} /* circ_flag */

if (KW_flag) {
  if (cw_node_in<=cw_node_out) dir=1; else dir=-1;
  i=cw_node_in;
  do { /* walk thru nodes */
    if (i==cw_node_in) Qi[i][4]=dQ_cw_in; 
    else               Qi[i][4]=-Qi[i-dir][5]; 
    MixPlug(&Qi[i][5],Qi[i][4],mp_cp_cw_h,mc_node,i,cw_node_in,cw_node_out,
            deriv_o,deriv_n,unit);
    qp_sum[unit][i][4]+=Qi[i][4];
    qp_sum[unit][i][5]+=Qi[i][5];
    i+=dir; 
  } while ( (min_cw<=i) && (i<=max_cw) ) ;
} /* KW_flag */


/* amb-losses & Qp-kAdown-up */
if (last_time_flag) {
  /* loop from n-1 to 1 down-up */
  dir=-1;
  i=n-1;
  if (i<1) { printf("storage[%d] n<=1!!! \n",unit); exit (-1); }
  Qi[n][2]=0;
  do { /* walk thru nodes */
       /* No coefficient for heat conduction given: Use
          Lambda(Water 20C) = 0.598W/Km, must be multiplied with cross 
          section area and divided by distance over which heat is conducted: */
    if (kA_heat_conduction==0) {
       if (StorageHeight!=0) Q_help=0.598*h*( deriv_o[unit][i+1]-deriv_o[unit][i]  )
                  * Vol/(1000.0*StorageHeight*StorageHeight/( (double) n));
       else  /* Use standard storage height of 2.0 m */
          Q_help=0.598*h*(deriv_o[unit][i+1]-deriv_o[unit][i])
                  * Vol/(1000.0*2.0*2.0/( (double) n));
#if 0
             if (StorageHeight!=0)
             Q_help=0.598*h*( deriv_o[unit][i+1]-deriv_o[unit][i]  )
                  * Vol/1000.0/StorageHeight/ (StorageHeight/( (double) n));
             else  /* Use standard storage height of 2.5 m */
             Q_help=0.598*h*( deriv_o[unit][i+1]-deriv_o[unit][i]  )
                  * Vol/1000.0/2.5 / (2.5/( (double) n));
            }
#endif
    }
    else Q_help=kA_heat_conduction*(deriv_o[unit][i+1]-deriv_o[unit][i])*h; /* given value */
    if (i==n) Q_help=0; /* deriv[n+1] not defined */
    if (    (deriv_o[unit][i+1]>deriv_o[unit][i]) && ( i!=n ) ) {
      /* mixing nodes */
      if (kA_down_up==0) Q_help+=0.5*mc_node*(deriv_o[unit][i+1]-deriv_o[unit][i]);
      else               Q_help+=kA_down_up*(deriv_o[unit][i+1]-deriv_o[unit][i])*h;  
    }
    Qi[i]  [2]  = Q_help;
    Qi[i+1][2] -= Q_help;
    qp_sum[unit][i+1][2]+=Qi[i+1][2];
    i+=dir; 
  } while ( (1<=i) && (i<=n) ) ;
  qp_sum[unit][1][2]+=Qi[1][2]; /* finish qp_sum */
  if (0) for (i=1;i<=n;i++) printf("Q2[%d]:%.1f\n",i,Qi[i]  [2]);

  /* ambient losses */
  if (StorageHeight>0) kA_node_corr=kA_node*1.1; /* + 10% */
  else kA_node_corr=kA_node;
  dir=-1;
  i=n;
  do { /* walk thru nodes */
    if (StorageHeight>0) {
      if ( (i==1) || (i==n) ) kA_node_corr=kA_node*1.1; /* + 10% */
      else                    kA_node_corr=kA_node*(1-(2*0.1/( (double) (n-2) )) );
    }
    else kA_node_corr=kA_node;
    Qi[i][3]=-kA_node_corr*h*(deriv_o[unit][i]-Tamb); /* last node extra */
    qp_sum[unit][i][3]+=Qi[i][3];
    i+=dir; 
  } while ( (1<=i) && (i<=n) ) ;
} /* last_time_flag */


/* ----------------------- energy balance ------------------------------- */
if ( (a)&&(last_time_flag) ) {
  printf("tsim:%.1f\n",tsim);
  for (i=1;i<=n;i++) { 
    printf("%d: ",i);
    for (i1=1;i1<=8;i1++) printf("Q[%d]:%.1f ",i1,qp_sum[unit][i][i1]);
    printf("\n");
  }
}

if (last_time_flag) {
  for (i=1;i<=n;i++) {
    energy[storage_counter][i]=init_energy[unit][i];
    for (i1=1;i1<=8;i1++) energy[storage_counter][i]+=qp_sum[unit][i][i1];
    deriv_n[unit][i]=energy[storage_counter][i]/mc_node;
  }
}


/*--outs----------------------------------------------------------------------*/
if (call_HE2_flag) {
  out_n[unit][1]=Qout_HE2;
  out_n[unit][2]=out_n IN (2);
}
if (KW_flag) {
   out_n[unit][3]=-Qi[cw_node_out][5]; /* Q_WW_out */
   out_n[unit][4]=out_n IN(4);
   P_WW=(out_n[unit][3]-out_n IN(3) )/h; 
}

/*System Energy */
/*if ( (last_time_flag[storage_counter]==select_unit[unit]) && !init_flag ) { */
if (last_time_flag) {
 out_n[unit][6]=0;
 out_n[unit][7]=0; for (i=1;i<=n;i++) out_n[unit][7]+=-Qi[i][3]; /* Q_loss amb */
 out_n[unit][8]=0; 
 for (i=1;i<=n+n_pipe1+n_pipe2;i++) out_n[unit][8]+=energy[storage_counter][i];
 if (a) printf("tsim:%.1f call:%d 7:%.1f 8:%.1f \n",tsim,select_unit[unit]
               ,out_n[unit][7],out_n[unit][8]);
}
else if (a) printf("else: tsim:%.1f call:%d 7:%.1f 8:%.1f \n",tsim
                   ,select_unit[unit],out_n[unit][7],out_n[unit][8]);

if (call_HE1_flag) {
  out_n[unit][9] =Qout_HE1;
  out_n[unit][10]=out_n IN(6);
}
  /* T_WW */
out_n[unit][11]=deriv_n[unit][ node_sensor1 ]; /* Tsensor1: node 1 T_WW */
if (out_n[unit][11]>130) { printf("storage[%d] T1:%.1f gau \n"
     ,unit,out_n[unit][11]); exit (-1); }
out_n[unit][12]=deriv_n[unit][ node_sensor2 ]; /* Tsensor2 */
out_n[unit][13]=deriv_n[unit][ node_sensor3 ]; /* Tsensor3 */
out_n[unit][14]=deriv_n[unit][ node_sensor4 ]; /* Tsensor4 */

  /* T of layer HE1_in  */
out_n[unit][15]=0;
  /* T_of layer HE2_in  */
out_n[unit][16]=0;

if (call_HE2_flag) out_n[unit][20]=(out_n IN (1)-out_n[unit][1])/h; /* P_HE2  [W] */
if (KW_flag) out_n[unit][22]=P_WW;
/*
if (KW_flag) printf("tsim:%.1f P_WW:%.1f mp:%.1f Tin:%.1f\n"
,tsim,out_n[unit][22],out_n[unit][4]*3600,out_n IN(3)/mp_cp_cw_h  );
*/

if (circ_flag) { /* Q_circ_out */
  out_n[unit][17]=-Qi[cw_node_out][7]; /* Q_circ_out=par23 */
  out_n[unit][18]=out_n IN (8);
  out_n[unit][27]=(-out_n[unit][17]+out_n IN (7) )/h; /*Power of circ */
}
  
if (call_HE1_flag) out_n[unit][21]=(out_n IN(5) -out_n[unit][9])/h;   /* P HE1 */

if (last_time_flag) { /* [W] Power of losses to ambient */
  out_n[unit][23]=0; for (i=1;i<=n;i++) out_n[unit][23]-=Qi[i][3];
  out_n[unit][23]/=h;
}         
  
if (call_HE2_flag) out_n[unit][24]=ToutHE2; /* temp HE2out (out3 of pipe2) */

if (last_time_flag) {  /* Tmean */
  out_n[unit][25]=0; for (i=1;i<=n;i++) out_n[unit][25]+=deriv_n[unit][i];
  out_n[unit][25]/=( (double) n );
  out_n[unit][26]=0; /*x !!!!!!! */
  for (i=6;i<=n;i++) out_n[unit][26]+=energy[storage_counter][i];
}
/*out27: P_circ:earlier set at circ flag */
if (call_HE2_flag) out_n[unit][28]=TinHE2;  /* temp HE2 inlet */
if (call_HE1_flag) {
 out_n[unit][29]=TinHE1;  /* temp HE1 inlet */
 out_n[unit][30]=ToutHE1;  /* temp HE1 outlet*/
}

  /* node temperature */
if (last_time_flag) {
    if ((30+n+n_pipe1+n_pipe2)>OUT_MAX0)
       {printf("storage n>n_max\n"); exit (-1); }
    for (i=1;i<=n+n_pipe1+n_pipe2;i++) out_n[unit][30+i]=deriv_n[unit][i];
/*
    printf("tsim:%.1f=%.2fh Tset=45;T1:%.1f PWW:%.1f KA_HE1:%.1f dTWT:%.2f TIn:%.1f \n"
      ,tsim,tsim/3600,deriv_n[unit][1],out_n[unit][22],kA_l_HE1*l1,
     -0.5*(deriv_n[unit][3]+deriv_n[unit][4])
     +0.5*(out_n[unit][29]+out_n[unit][30]),out_n[unit][29] );
*/
}

if (a) for (i=1;i<=n+n_pipe1+n_pipe2;i++) printf("T%d:%.1f ",i,deriv_n[unit][i]); 
if (a) printf("Tin:%.1f THE1_1,2,3:%.1f,%.1f,%.1f mp_HE1:%.1f\n",
   TinHE1,deriv_n[unit][11],deriv_n[unit][12],deriv_n[unit][13],3600*mp_HE1);

  if ( (last_time_flag) && (pedantic)  ) { /* unit energy check */
    static double Qunit_transfer_sum[STORAGE_MAX]={0,0,0,0,0,0,0,0},
                  Qinit[STORAGE_MAX]={0,0,0,0,0,0,0,0},
                  Qsys_losses_sum[STORAGE_MAX]={0,0,0,0,0,0,0,0};
    double Qunit_transfer,Qsys_losses,energy_change;

    energy_change=out_n[unit][8]-out_o[unit][8];
    Qunit_transfer= out_n IN (1)-out_n[unit][1]+out_n IN (3)-out_n[unit][3]
                   +out_n IN (5)-out_n[unit][9]+out_n IN (7)-out_n[unit][17];
    Qsys_losses=out_n[unit][7];
    if ( ( fabs( Qunit_transfer-Qsys_losses-energy_change)>1 ) && (!init_flag) ) {
      printf("storage%d: energy_check error tsim:%.1f Qunit_transfer:%.1f "
             "Qsys_losses:%.1f energy_change:%.1f error:%.1f \n",
             unit,tsim,Qunit_transfer,Qsys_losses,energy_change,
             Qunit_transfer-Qsys_losses-energy_change);
      printf("in1:%.1f out1:%.1f in3:%.1f out3:%.1f in5:%.1f out9:%.1f in7:%.1f out17:%.1f\n",
        out_n IN (1),out_n[unit][1],out_n IN (3),out_n[unit][3],out_n IN (5),out_n[unit][9],
        out_n IN (7),out_n[unit][17]);
      printf("mp1in:%.2f mp1out:%.2f\n",out_n IN (2)*3600,out_n[unit][2]*3600);
      exit (-1);
    } /* error */
    /* history energy check */ 
    if (init_flag) {
      Qinit[storage_counter]=out_n[unit][8]+Qsys_losses-Qunit_transfer;
      Qsys_losses_sum[storage_counter]=Qsys_losses;
      Qunit_transfer_sum[storage_counter]=Qunit_transfer;
      }
    else {
     Qunit_transfer_sum[storage_counter]+=Qunit_transfer;
     Qsys_losses_sum[storage_counter]+=Qsys_losses;
    }
    if ( ( fabs( Qinit[storage_counter]+Qunit_transfer_sum[storage_counter]
           -Qsys_losses_sum[storage_counter]-out_n[unit][8])>2 ) && (!init_flag) ) {
      printf("storage: history energy_check error tsim:%.1f Qinit:%.1f Qunit_transfer_sum:%.1f "
             "Qsys_losses_sum:%.1f sys_energy:%.1f error:%.1f\n",
             tsim,Qinit[storage_counter],Qunit_transfer_sum[storage_counter],
             Qsys_losses_sum[storage_counter],out_n[unit][8],
             Qinit[storage_counter]+Qunit_transfer_sum[storage_counter]
               -Qsys_losses_sum[storage_counter]-out_n[unit][8]);
      printf("in1:%.1f out1:%.1f in3:%.1f out3:%.1f in5:%.1f out9:%.1f in7:%.1f out17:%.1f\n",
        out_n IN (1),out_n[unit][1],out_n IN (3),out_n[unit][3],out_n IN (5),out_n[unit][9],
        out_n IN (7),out_n[unit][17]);
      exit (-1);
    } /* error */
  } /* unit energy check */
  return 1;
}

int internal_pipe(double *dQout,double *dQloss,double dQin,double mp_cp_h,
                  int current_node,int i_offset,double Tamb,double kA,
                  double mc_i,double ***qp_sum,int unit,
                  double energy[STORAGE_MAX][DERIV_MAX],int index,
                  double **deriv_o,double **deriv_n,double **init_energy,
                  double h)
{
int a=0;
int i1;
double energy_with_plug;
double T_tmp,deltaT;
int pedantic=0; /* only one storage energy balance check */

i1=current_node+i_offset;
energy_with_plug=dQin+mc_i*deriv_o[unit][i1]; /* get new mixed node */
T_tmp=energy_with_plug/(mp_cp_h+mc_i); /* get temperary temp of node */
deltaT=T_tmp-Tamb;
*dQloss=deltaT*kA*h;
T_tmp=(energy_with_plug- *dQloss)/(mp_cp_h+mc_i); /* with amb losses */
if (   ( (*dQloss>0) && (T_tmp<Tamb) ) /* instable */
    || ( (*dQloss<0) && (T_tmp>Tamb) )
   ) {
   if (pedantic) printf("internal_pipe[%d]:stabilization T_tmp:%.1f Tamb:%.1f Qi:%.1f \n",
                        unit,T_tmp,Tamb,*dQloss);
   *dQloss=energy_with_plug-Tamb*(mp_cp_h+mc_i);/* set Tnode to Tamb */
}
*dQout=mp_cp_h*T_tmp;

qp_sum[unit][i1][1]+=dQin;
qp_sum[unit][i1][2]+=*dQloss;
qp_sum[unit][i1][3]+=*dQout; 
energy[index][i1]=init_energy[unit][i1]
                 +qp_sum[unit][i1][1]-qp_sum[unit][i1][2]-qp_sum[unit][i1][3];
deriv_n[unit][i1]=energy[index][i1]/mc_i;
if (a) printf("internal_pipe: i:%d index:%d deriv_n:%.1f\n",i1,index,deriv_n[unit][i1]);
return (1);
} /* internal_pipe() */

int MixPlug(double *dQout,double dQin,double mp_cp_h,double mc,int current_node,
            int in_node,int out_node,double **deriv_o,double **deriv_n,int unit) 
{
/* i=current_node;in_node=dQ_circ_in;out_node=cw_node_out */
/* mix in-plug with node-cap mc, shift out Qout(Tmixed) */
int a=0;
int i=current_node;
double mc_mix,Qmix,T_tmp;

mc_mix=mc+mp_cp_h;
Qmix=dQin+mc*deriv_o[unit][i];
T_tmp=Qmix/(mc_mix);
*dQout=-mp_cp_h*T_tmp;
if (mp_cp_h<1e-6) *dQout=0;
if (a) printf("###########MixPlug: T_tmp:%.1f i:%d dQin:%.1f mp_cp_h:%.1f\n"
              ,T_tmp,current_node,dQin,*dQout);
return (1);
}


int GetStratisfiedInletNode(double dQin,double mp_cp_h,int n,int unit,double **deriv_o) 
{
/* call:  HE2_in=GetStratisfiedInletNode(Tin,n,unit,deriv_o); */
int node_in;
double Tin;
int i;

Tin=dQin/(mp_cp_h+1e-12);
if ( (Tin>150) || (Tin<-5))  printf("storage[%d] Tin:%.1f \n", unit,Tin);
i=1; /* upper node */
while ( (Tin<deriv_o[unit][i]) && (i<n) ) i++;
node_in=i;
return (node_in);
}
