/*  global.orig.h  ColSim,Christof Wittwer FhG-ISE FBTA                       */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/
/* this file is used to generate a global.h with cfg/config routine */

/*
000530 horizon prototype added
000607 pointer lenght,latitude
000609 type_call with lenght and latitude
001109 dek_reader_header out_label, type_call header
001121 double Get_phi_from_xs(double x,double x_s); deklaration
*/

#define WIN32 0  /* 0: compile for UNIX 1: compile for win32  */

#define INPUTMAX  10     /* fuzzy_controller */

#define MAX_TYPES  200    /* Max number of units */

#define UNIT_MAX  200    /* Max units + DEAD_MAX */
#define PAR_MAX   70 
#define INPUT_MAX 40
#define OUT_MAX0  50 /* number of outputs (excluding init_values of inputs) */
#define OUT_MAX OUT_MAX0+INPUT_MAX  /* OUT_MAX0+INPUT_MAX
                                    OUT_MAX0+1: input initial 1
                                    OUT_MAX0+2: input initial 2 , ... etc...
                                          Input initials for out(>OUT_MAX0)*/

#define PATH_MAX 10            /* No of Energy Flows to Node */
#define MAX_POINTS 300        /* Max Number of values for t-Axes */

#define STORAGE_MAX 10
#define DEAD_TIME_MAX 10000     /* max no of layers (*h) */
#define DEAD_MAX 40           /* max no of dead_time-components */
                              /* Attention: collect. needs 4 comp's */
                              /*            storage:2 wt's          */
#define PUMP_NR 20    /*max number of pumps */
#define DIV_NR 30     /* max number of flow diverters */
#define NODE_MAX 101   /* max no of nodes in one TYPE, array size of **Qi  */
/* #define DERIV_MAX INPUT_MAX     earlier:15, bug in printer 10/97 */
#define DERIV_MAX NODE_MAX    /* earlier:15, bug in printer 10/97 */

/*
 #define IN 0
 #define OUT 1        point_dead[][OUT] at outlet temp  
*/

#define WEATHER_SIZE 10000  /* lines in weather[] */
#define LOAD_SIZE 5000       /* lines in load[]  */

#define NR_TETA_MAX 91 /* iam table  */
#define NR_PHI_MAX 361
#define XXX       -999

#define TYPES_STRING_LENGTH  13 /* name for the types */


/* Type Definitions:
  type es explicit defined in main(), init_type_def_str.c */

#define TypeAirRadNode 29   /* Now used in wall.c */ 


/* some constants */
#define PI   3.1415926535897932384626433832795029 /* see M_PI in math.h */
#define CP_AIR_DRY 1007    /* [J/kgK] spec. heat capacity dry air (p=ct) */
#define CP_STEAM   1860    /* [J/kgK] spec. heat capacity steam (water)*/
#define CP_WATER   4190    /* [J/kgK] spec. heat capacity water */
#define EVAP_ENTH  2501000 /* [J/kgK] Evaporation enthalpy water */
#define P_ATM 1000         /* [hPa]  hPa=100 Pa = mbar */

/* some useful things */
/* reading inputs
   Q_in=out_n[  in_u[unit][1] ][ in_nr[unit][1] ];
   Q_in=out_n IN_01;
*/
#define PAR(x)   par[unit][x] 
#define IN(x)    [in_u[unit][x]][in_nr[unit][x]] /* macro */
#define OUT(x)   out_n[unit][x]
#define DERIV(x) deriv_n[unit][x]


/* -------------malloc_call.c--------------------------------------*/
char**    malloc_2char(int x,int y);
char***   malloc_3char(int x,int y,int z);
double*   malloc_1double(int x);
double**  malloc_2double(int x,int y);
double*** malloc_3double(int x,int y,int z);
int**     malloc_2int(int x,int y);
int*      malloc_1int(int x);


/* ColSim history
double deadtime(int nr2,int t_dead,double Q_in,double **Q,
                int **point_dead,double mp_cp_h,double m_rest_cp
                ,double energy_dead, double mc);  

double deadtime_change(int nr3,int t_dead_old,int t_dead_new,double **Q
                       ,int **point_dead,double mc,double cp
                       ,double energy_dead
                       ,double mp_cp_h_old,double m_rest_cp_old
                       ,double mp_cp_h    ,double m_rest_cp    );
*/

double energy_outcome(int unit,double **out_n,double tsim,
                      int init_flag,int stop_flag,double h,int un_max,
                      double **init_energy,double **out_o,int *type,
                      char **type_def_str,double max_error,
                      double ***qp_sum,
		      double **deriv_o,double **deriv_n,
		      int **in_u,int **in_nr);

/*-----------------------------------------------------------------*/
int init_set_mp(double **par,int **in_u,int **in_nr,double **out_o,
           double **out_n,int *type,int pump_nr,int *pump_unit,
           int coll_pipe_nr,int *coll_pipe_unit,double tsim,
           int un_max,char **type_def_str,
           int i_div,int *div_unit,int mp_change,int init_flag,
           int **active_circle);

int init_type_def_str(char **type_def_str);

/*-----------print_infos.c-----------------------------------------*/
int Ausgabe_Input(int **in_u,int **in_nr,int un_max,
                  int *type,char **type_def_str);

int Ausgabe_Out_pr(double **out_n,double **out_o,int un_max,
                   int *type,char **type_def_str, int difference );

int dek_reader(int *type,double **par,int **in_u,int **in_nr,
               double **out_n,double **out_o,
               double **deriv_o,double **deriv_n,char ***out_label);


/*------------------------------weather.c--------------------------*/
/* local definition
double weather_init(int unit,double **out_o,double **out_n,
                    int columns,double *weather,int year_start,int month,
                    int day,int unix_flag,int *weather_flag,double **par);
*/
/*------------------------------------------------------------------*/

/*------------------------------load.c-TYPE 14 --------------------*/
double load_init(int unit,double **par,double **out_o,
                 double **out_n,double **load);

/*-----------------------------------------------------------------*/
int rad_processor(double tsim,int day_of_year,double Ib_hor, double Id_hor
                  ,double rho_g,double E_40,double slope,double orientation
                  ,double *out,double *lenght,double *latitude,double Rd
                  ,double Rr);
int rad_proc2(double *Ibt,double *Idt,double *Igt,double *out,
              double Igh,double Idh,double Ibn,double Igt_mess,
              int mode,int init_flag,double tsim,double h,int day_of_year,
              int horizon_nr,double b0,
              double slope,double orientation,double *lenght,double *latitude,
              double rho_g);

/*-----------------angle_modifier-iam----- ------------------------*/
double inc_angl_mod(int init_flag, double b0, double teta,double phi
                    ,double **iam);
double inc_angl_mod2(int init_flag, double b0, double teta,double phi);

double  horizon(int init_flag,double teta_z,double azimut /*gamma_s */
           ,double *Rd,/*ratio_diffus*/ double *Rr /* ration reflect */
           ,double albedo,double slope,double orientation);


/*---------------- defs for thermodynamics --------------------------------*/
double Get_phi_from_xs(double x,double x_s);
double Get_xs_from_enthalpy(double H,double m_air,double m_water,double T);
double Get_xs_from_enthalpy2(double h); /* christiane */

double Get_Temp_from_enthalpy(double H,double m_air,double m_steam
                             ,double m_water);
double Get_p_steam_s (double T_air);       /* [N/m2] pressure saturated air*/
double Get_phi(double T_air,double x);

double Get_x (double p_steam, double p_air);
double Get_rho_air_dry (double T_air, double p_air);
double Get_rho_humid (double T_air, double x, double p_air);
double Get_rho_co2(double T_air);
double Get_h_air_dry (double T_air);         /*[Ws/kg] enthalpie dry air*/
double Get_h_steam (double T_air);        /*[Ws/kg] enthalpie water steam*/
double Get_h_water (double T_air);         /*[Ws/kg] enthalpie water*/


double Get_alpha_air (double T_air, double di, double l_p, double m_in,
                    double t_step, double rho);
                                          /*[W/m2K] heat transfer air-pipe*/
double Get_alpha_water (double T_water, double di, double l_p, double m_in,
                      double t_step);       /*[W/m2K] heat transfer water-pipe*/
double Get_f_mass (double a, double lambda, double alpha, double di, double da);

double Get_alpha_air (double T_air, double di, double l_p, double m_in,
                    double t_step, double rho);
                                           /*[W/m2K] heat transfer air-pipe*/
double Get_alpha_water (double T_water, double di, double l_p, double m_in,
                      double t_step);     /*[W/m2K] heat transfer water-pipe*/
double Get_f_mass (double a, double lambda, double alpha, double di, double da);
double Get_H_total (double T,double m_dry,double m_steam,double m_water)  ;
double Get_h_total (double T,double m_dry,double m_steam)  ;
 
 
int unit_order_init(
           int    unit,      /* current unit number */
           double **par,     /* parameter array par[unit][1..PARMAX]  */
           int    **in_u,    /* input unit array, input connections   */
           int    **in_nr,   /* input number array                    */
           double **out_o,   /* output array with privious state      */
           double **out_n,   /* output array with current state       */
           double **deriv_o, /* derivatives: used for diffential equ. */
           double **deriv_n, /*  new state of deriv                   */
           double h,         /* sim time step                         */
           double **init_energy, /* initial energy in RC-comp. [Ws]   */
           double ***qp_sum, /* energy to one node qp[unit][n][m]     */
           double tsim,      /* simulation time in [s] of this day    */
           int    init_flag, /* set at initial loop of sim            */
           char   **type_def_str,
           int    *type,
           int    unit_max,  /*            */
           int    *unit_order
 );




int type_call(
              char **type_def_str,
              int unit,
              int un_max,
              int *type,
              int *select_unit,
              int *type_index,
              int init_flag,
              double *lenght,
              double *latitude,
              int day_of_year,
              int year_start,
              int month,
              int day,
              int hour,
              int min,
              int sec,
              int day_flag,
              int hour_flag,
              int week_flag,
              int day_of_week,
              int min10_flag,
              int stop_flag,
              int previous_unit,
              int **in_u,
              int **in_nr,
              int **point_dead,
              double **iam,
              double h,
              double tsim,
              double **par,
              double **out_n,
              double **out_o,
              double **deriv_o,
              double **deriv_n,
              double **init_energy,
              double **Qi,
              double **Q, /* pipe */
              double ***qp_sum,
              int a,
              int *unit_order,
              int nr_of_unit_calls,
              double **dummy,
              char ***out_label
             );

