/*  global.orig.h  ColSim,Christof Wittwer FhG-ISE FBTA                       */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/
/* this file is used to generate a global.h with cfg/config routine */

/*
000530 horizon prototype added
000607 pointer lenght,latitude
000609 type_call with lenght and latitude
001109 dek_reader_header out_label, type_call header
001121 double Get_phi_from_xs(double x,double x_s); deklaration
010517 unit_order_init() par first_model_index added
010612 chriwi:Get_enthalpy_from_temp declared
010522 INPUTMAX fuzzy erased
010627 skip_time_intervall,**thermo_struct added to type_call()
       malloc struct added
010920 timezone added to args of type_call
011016 rad_processor.h included
020404 MAX_POINTS moved to gnuplotter.c
020526 LOAD_SIZE moved to load.c
*/


#define WIN32 0  /* 0: compile for UNIX 1: compile for win32  */

/* #define INPUTMAX  10     / * fuzzy_controller */

#define MAX_TYPES  200    /* Max number of units */

#define UNIT_MAX  200    /* Max units + DEAD_MAX */
#define PAR_MAX   70 
#define INPUT_MAX 40
#define OUT_MAX0  50 /* number of outputs (excluding init_values of inputs) */
#define OUT_MAX OUT_MAX0+INPUT_MAX  /* OUT_MAX0+INPUT_MAX
                                    OUT_MAX0+1: input initial 1
                                    OUT_MAX0+2: input initial 2 , ... etc...
                                          Input initials for out(>OUT_MAX0)*/

#define PATH_MAX 10            /* No of Energy Flows to Node */
/* #define MAX_POINTS 300    Max Number of values for t-Axes */

#define STORAGE_MAX 10
#define DEAD_TIME_MAX 10000     /* max no of layers (*h) */
#define DEAD_MAX 40           /* max no of dead_time-components */
                              /* Attention: collect. needs 4 comp's */
                              /*            storage:2 wt's          */
#define PUMP_NR 20    /*max number of pumps */
#define DIV_NR 30     /* max number of flow diverters */
#define NODE_MAX 101   /* max no of nodes in one TYPE, array size of **Qi  */
/* #define DERIV_MAX INPUT_MAX     earlier:15, bug in printer 10/97 */
#define DERIV_MAX NODE_MAX    /* earlier:15, bug in printer 10/97 */


#define WEATHER_SIZE 10000  /* lines in weather[] */

#define NR_TETA_MAX 91 /* iam table  */
#define NR_PHI_MAX 361
#define XXX       -999

#define TYPES_STRING_LENGTH  13 /* name for the types */
#define TYPES_MAX            100 /* max number of types */


/* Type Definitions:
  type es explicit defined in main(), init_type_def_str.c */

#define TypeAirRadNode 29   /* Now used in wall.c */ 


/* some constants */
#define PI   3.1415926535897932384626433832795029 /* see M_PI in math.h */
/* defines fluid properties */
#define METHOD_WATER   1
#define METHOD_GLYCOL  2

#if 0
#define CP_AIR_DRY 1007    /* [J/kgK] spec. heat capacity dry air (p=ct) */
#define CP_STEAM   1860    /* [J/kgK] spec. heat capacity steam (water)*/
#define CP_WATER   4190    /* [J/kgK] spec. heat capacity water */
#define EVAP_ENTH  2501000 /* [J/kgK] Evaporation enthalpy water */
#define P_ATM 1000         /* [hPa]  hPa=100 Pa = mbar */
#endif

/* define input,output and parameter macro: 
   Q_in=out_n[  in_u[unit][1] ][ in_nr[unit][1] ];
   --> Q_in=out_n IN (1)
*/
#define PAR(x)   par[unit][x] 
#define IN(x)    [in_u[unit][x]][in_nr[unit][x]] /* macro */
#define OUT(x)   out_n[unit][x]
#define DERIV(x) deriv_n[unit][x]


/* -------------malloc_call.c-------------------------------------------------*/
/* allocation routines */
struct    therm_hydraulic_struct* malloc_thermo_struct(int index);
float *malloc_1float(int x);
char**    malloc_2char(int x,int y);
char***   malloc_3char(int x,int y,int z);
double*   malloc_1double(int x);
double**  malloc_2double(int x,int y);
double*** malloc_3double(int x,int y,int z);
int**     malloc_2int(int x,int y);
int*      malloc_1int(int x);


/* global structs which belong to the energy calculation */
/* and provide necessary infos tor types */




/* thermo_struct declared global */
/* thermo_hydraulics mh:
rho    density  (Dichte in [kg/m^3])
ny      viscosity : (kinematische Viskositaet in [m^2/s])
cp     (spezifische Waermekapazitaet in [kJ/(kg*K)]
lambda (Waermeleitung in [W/(m*K)]

results of hydraulic solve:
vol   volume_flow (Volumenstrom in [m^3/h])
dp_a  delta_p_path_a (Druckverlust im abzweigenden Pfad in [Pa]; enthaelt nur bei 
       Diverter und Mixer einen Wert ungleich 0)
dp_d  delta_p_path_d (Druckverlust im durchgehenden Pfad in [Pa]; enthaelt z. B. auch 
       den Druckverlust eines Rohres)
ctr   ratio_flow_path_ad  Volumenstromverhltnis (Diverter/Mixer)
future idea of this struct:
interface[unit].flowout
interface[unit].flowin
state[unit]:
  state[unit].fluidmasses.dm_air
  state[unit].fluidmasses.dm_steam
  state[unit].fluidmasses.dm_fluid
*/

struct sim_struct { /* relevant parameters for simulation process */
 int unit; /* unique ID for the objects */
 int type; /* model type */
 char info[12]; /* model info string */
 int init_flag; /* =1, if sim-start (tsim=tstart) */
 double h; /* sim time step */
 double tsim; /* [s] sim time of the day */
 int day_of_year;
 double previous_unit; /* this unit was processed just before */
 int active_plug_input; /* current plug is pushed to this input */
 int select_unit; /* number of call of this unit due to this time step */
 int type_index; /* index for same types, internal array index */
 int last_time_flag; /* last call of this unit due to this timestep */
}; /* definition in main() : struct sim_struct info */



struct interface_struct { /* complete interface and state */
 double* parameter; /* e.g. interface[unit]->parameter[1] */
 int*    connected_unit; /* **in_u */
 int*    out_number_of_connected_unit; /* **in_nr */
 double* out_current; /* **out_n */
 double* out_last_timestep; /* **out_o; */
 double* state_current; /* **deriv_n */
 double* state_last_timestep; /* **deriv_o */
 double  init_energy;
 double  sys_gains; /* losses are negative */
 double  sys_energy; /* internal energy of this unit */
 double  m_gains; /* mass_gains regarding to this system */
 double  m_flow_internal; /* mass_balance of flow */
}; /* definition in main() struct interface_struct interface ; */

struct radiation_struct {
 int    mode; /* type of rad_processingr,.e.g IGH_MES */
 int    horizon_nr;
 double b0;
 double iam_diff;
 double slope;
 double orientation;
 double rho_g;

 double timezone;
 double length;
 double latitude;

 double Ibt;
 double Idt;
 double Igt;
 double Ibt_trans; /* with transmission thru glacing and blinds */
 double Idt_trans;
 double Igt_trans;
 double Igh;
 double Idh;
 double Ibn;
 double Igt_mess;
 double theta_z;
 double gamma_s;
 int    shading_unit;
 int    wall_unit;
 int    mode_shading;
 double rho_shading;
 double Rd;
 double Rr;
 double inc_angl;
 double g;
 double tau; /* tau transmission  */
 double theta;
}; /* definition in main()  struct radiation_struct radiation */




struct therm_hydraulic_struct {
 int    type;
 int    pump_unit;
 int    fluid_property_class;
 double density;
 double viscosity;
 double cp;
 double lambda;
 double volume_flow;
 double delta_p_path_a;
 double delta_p_path_d;
 double ratio_flow_path_ad;
 double m_air;
 double m_fluid;
 double m_steam;
 double m_co2;
};

/*  fluid_properties.c */
double Get_cp_from_Temp(int unit,int method,double T);
double Get_rho_from_Temp(int unit,int method,double T);

double energy_outcome(int unit,double **out_n,double tsim,
                      int init_flag,int stop_flag,double h,int un_max,
                      double **init_energy,double **out_o,int *type,
                      char **type_def_str,double max_error,
                      double ***qp_sum,
		      double **deriv_o,double **deriv_n,
		      int **in_u,int **in_nr);

/*-----------------------------------------------------------------*/
int init_set_mp(double **par,int **in_u,int **in_nr,double **out_o,
           double **out_n,int *type,int pump_nr,int *pump_unit,
           int coll_pipe_nr,int *coll_pipe_unit,double tsim,
           int un_max,char **type_def_str,
           int i_div,int *div_unit,int mp_change,int init_flag,
           int **active_circle);

int init_type_def_str(char **type_def_str);

/*-----------print_infos.c-----------------------------------------*/
int Ausgabe_Input(int **in_u,int **in_nr,int un_max,
                  int *type,char **type_def_str);

int Ausgabe_Out_pr(double **out_n,double **out_o,int un_max,
                   int *type,char **type_def_str, int difference );

int dek_reader(int *type,double **par,int **in_u,int **in_nr,
               double **out_n,double **out_o,
               double **deriv_o,double **deriv_n,char ***out_label);



/*------------------------------load.c-TYPE 14 -------------------------------*/
/* declaration of load.c is autoconfigured by cfg/config */
double load_init(int unit,double **par,double **out_o,
            double **out_n,double **load,double skip_time_intervall,
            int lines_max);

/*-----------------angle_modifier-iam----- ------------------------*/
double inc_angl_mod(int init_flag, double b0, double teta,double phi
                    ,double **iam);
double inc_angl_mod2(int init_flag, double b0, double teta,double phi,
                     double *iam_diff);

double  horizon(int init_flag,double teta_z,double azimut /*gamma_s */
           ,double *Rd,/*ratio_diffus*/ double *Rr /* ration reflect */
           ,double albedo,double slope,double orientation);


/*---------------- defs for thermodynamics --moved to thermodynamics.h--------*/
#if 0
double Get_enthalpy_from_temp(double T,double m_air,double m_steam,
                              double m_water);

double Get_phi_from_xs(double x,double x_s);
double Get_xs_from_enthalpy(double H,double m_air,double m_water,double T);
double Get_xs_from_enthalpy2(double h); /* christiane */

double Get_Temp_from_enthalpy(double H,double m_air,double m_steam
                             ,double m_water);
double Get_p_steam_s (double T_air);       /* [N/m2] pressure saturated air*/
double Get_phi(double T_air,double x);

double Get_x (double p_steam, double p_air);
double Get_rho_air_dry (double T_air, double p_air);
double Get_rho_humid (double T_air, double x, double p_air);
double Get_rho_co2(double T_air);
double Get_h_air_dry (double T_air);         /*[Ws/kg] enthalpie dry air*/
double Get_h_steam (double T_air);        /*[Ws/kg] enthalpie water steam*/
double Get_h_water (double T_air);         /*[Ws/kg] enthalpie water*/


double Get_alpha_air (double T_air, double di, double l_p, double m_in,
                    double t_step, double rho);
                                          /*[W/m2K] heat transfer air-pipe*/
double Get_alpha_water (double T_water, double di, double l_p, double m_in,
                      double t_step);       /*[W/m2K] heat transfer water-pipe*/
double Get_f_mass (double a, double lambda, double alpha, double di, double da);

double Get_alpha_air (double T_air, double di, double l_p, double m_in,
                    double t_step, double rho);
                                           /*[W/m2K] heat transfer air-pipe*/
double Get_alpha_water (double T_water, double di, double l_p, double m_in,
                      double t_step);     /*[W/m2K] heat transfer water-pipe*/
double Get_f_mass (double a, double lambda, double alpha, double di, double da);
double Get_H_total (double T,double m_dry,double m_steam,double m_water)  ;
double Get_h_total (double T,double m_dry,double m_steam)  ;
#endif
 
 
int unit_order_init(
           int    unit,      /* current unit number */
           double **par,     /* parameter array par[unit][1..PARMAX]  */
           int    **in_u,    /* input unit array, input connections   */
           int    **in_nr,   /* input number array                    */
           double **out_o,   /* output array with privious state      */
           double **out_n,   /* output array with current state       */
           double **deriv_o, /* derivatives: used for diffential equ. */
           double **deriv_n, /*  new state of deriv                   */
           double h,         /* sim time step                         */
           double **init_energy, /* initial energy in RC-comp. [Ws]   */
           double ***qp_sum, /* energy to one node qp[unit][n][m]     */
           double tsim,      /* simulation time in [s] of this day    */
           int    init_flag, /* set at initial loop of sim            */
           char   **type_def_str,
           int    *type,
           int    unit_max,  /*            */
           int    *unit_order,
           int    *first_model_index, /* FastIteration setting */
           int    *last_model_index   /* FastIteration setting */

 );



/* used in type_call.orig.c and type_call.c */
int type_call(
              struct sim_struct *info, /* see global.h */
              struct interface_struct *interface,
              struct radiation_struct *radiation,
              char **type_def_str,
              int unit,
              int un_max,
              int *type,
              int *select_unit,
              int *type_index,
              int init_flag,
              double *timezone,
              double *lenght,
              double *latitude,
              int day_of_year,
              int year_start,
              int month,
              int day,
              int hour,
              int min,
              int sec,
              int day_flag,
              int hour_flag,
              int week_flag,
              int day_of_week,
              int min10_flag,
              int stop_flag,
              int previous_unit,
              int **in_u,
              int **in_nr,
              double **iam,
              double h,
              double tsim,
              double **par,
              double **out_n,
              double **out_o,
              double **deriv_o,
              double **deriv_n,
              double **init_energy,
              double **Qi,
              double **Q, /* pipe */
              double ***qp_sum,
              int    a,
              int    *unit_order,
              int    nr_of_unit_calls,
              double **dummy,
              char ***out_label,
              double skip_time_intervall,
              struct therm_hydraulic_struct **thermo_struct
             );
