/*
   ColSim 0.1a
   System Simulation Program for Thermal Systems & Controllers
   Christof Wittwer/Fraunhofer-ISE/Universitaet Karlsruhe
   chrwi@ise.fhg.de, chriwi@fbta.uni-karlsruhe.de

   started at 95/01/01

   Thanks to following support:
   Ritter Umwelt GmbH and Paradigma Solar Systems
   ROM Umwelt Stiftung Hamburg
   RESOL GmBH Hattingen
   Fraunhofer-ISE, D-79100-Freiburg
   University Karlsruhe FBTA
   

   structure in ColSim:
   ./src           : makefile,source *.c
   ./load_profiles : load.xxx.dat
   ./cnv           : converter from xfig object
   ./macro         : shellscripts

   program start with   ./sim
   simulation script in ./sim.dek
   load_profiles in     ./load.dat  (./load_profiles/load.NAME.dat)
   output data   in     ./plot.dat for online plotter data
                        ./simxx.dat for simulation output data

 * Copyright (C) 1995  Christof Wittwer
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *

000425 dummy=NULL variable defined and add to type_call()
000727 if (select_unit[unit]==1) type_index[unit]=type_counter[  type[unit] ];
       bugfix type_index: 3rd storage-call with wrong type_index
001130 if debug flag set (sim -d ) , check NaN of all outputs at each simtime
010131 christiane: start_date can be later then stop date in order to calc
        e.g. heating period from autumn to spring.  
010517 FastItaration method added , 
       some syntac mod's, i_dead_nr erased
010522 SIMCONTROL parameters extended,skip_time_intervall is par8
       skip_timer ist loaded with skip_time_intervall, from tsim_start
010627 skip_time_intervall,**thermo_struct added to type_call()
010827 sim_struct added
010920 timezone added
010923 include "rad_processor.h"
010928 struct interface included ,point_dead erased 
020403 FastIteration_bak used at energy_outcome
020502 FastIteration updated new paremter , ommitting flipping values etc
020524 select_unit[i]=info_array[i]->select_unit=0; bug fix
*/

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "global.h"
#include "rad_processor.h"


double convergence(double **deriv_o,double **deriv_n,
                   double **out_o,double **out_n,
                   double eps_interrupt,
                   int i_iteration,int *unit_order,
                   int first_model_index,int last_model_index,int a);


int main(argc,argv)
 int   argc;
 char **argv;
{
int a=0; /*-----DEBUG with a=1----------------------*/

int laufunit,check; /* current_array*/
int i,i1;           /* different counters */
int un_max=0; /*max units in dek != UNIT_MAX*/

static char day_string[8][10]={
                         {"          "}, /* day_string[0][] */
                         {"Monday    "},
                         {"Tuesday   "},
                         {"Wednesday "},
                         {"Thursday  "},
                         {"Friday   "},
                         {"Saturday  "},
                         {"Sunday    "}
                                        };
static char month_string[25][9]={
                            {"         "},  /* month_string[0][] */
                            {"January  "},
                            {"February "},
                            {"March    "},
                            {"April    "},
                            {"May      "},
                            {"June     "},
                            {"July     "},
                            {"August   "},
                            {"September"},
                            {"Oktober  "},
                            {"November "},
                            {"December "},
                            {"January  "},
                            {"February "},
                            {"March    "},
                            {"April    "},
                            {"May      "},
                            {"June     "},
                            {"July     "},
                            {"August   "},
                            {"September"},
                            {"Oktober  "},
                            {"November "},
                            {"December "}
                                         };

static int no_days_of_month[25]={ 0,31,28,31,30,31,30,31,31,30,31,30,31,
                                 31,28,31,30,31,30,31,31,30,31,30,31};

int day_of_year=0; /* current day of the year */
int day_of_week=0; /* current day of the week */

/*----------------------------Sim-Start&Stop-----------*/

int first_day_of_year=6; /* 0=montag,6=sonntag */

int date_start=940701; /*year/month/day*/
int date_stop =940701;
int time_start=000000; /*hour/minute/sec*/
int time_stop =235959;
int unix_flag=1;       /* unix/os2 */

int i_month,j_month;   /* month-counter */
int year_start;       /* read year from sim_control */
int month_start;
int month_stop;
int i_day;                /* day-counter */
int day_start,day_start2;
int day_stop,day_stop2;
int i_hour;
int hour_start;
int hour_stop;
int i_min;
int min_start;
int min_stop;
int i_sec;
int sec_start;
int sec_stop;
int min10_flag;
int hour_flag;
int day_flag;
int week_flag;

int init_flag=1; /* set if first step, initilize */
int stop_flag =0; /* set if last step             */

double tsim=0;      /* current simulation time [s] of this day  */
double t_start=0.0;   /* 0*3600  [s] */
double t_stop=86399.0;  /* 23*3600+59*60+59;   sim end in [s] */
double h=1;/* time step of sim ,later set from par7,sim_control [s]*/

double max_error=1,dQ_error=0; /*System_Energie-Zuwachs,num. Fehler [Ws] */

int previous_unit=0; /* Q_inlet check */
int nr_of_unit_calls=0;    /*calls in timestep */

int mp_change=0; /* change of massflow in diverter or pump */

double *timezone;  /* 1= 1h east  */
double *lenght;
double *latitude;

int *type;
char **type_def_str;
double **parameter;
int **input_unit;
int **input_out_nr;
char ***out_label; /* get connection label from output comment */
int **active_circle; /* order of units set by set_mp() */
int *unit_order; /* [UNIT_MAX] order of units for simulation in timestep */
int *select_unit; /*flag for registration in active pump circle */
int first_model_index,last_model_index; /* FastIteration Mode */
int FastIteration=0,FastIteration_bak=0,NoOfIterations=1000; 
                                                      /* FastIteration Mode */
int i_iteration=1;
int stop_iteration=0;
int skip_flag=0; /* to skip model executing */
double skip_time_intervall=0; /* set from par8 simcontrol 30*60;  [s] */
double skip_timer=0;
double eps,eps_interrupt; /* error of deriv's */
int unit_order_start,unit_order_stop; /* loop definition */
 
int *type_index;   /* inc for each call of one type */
int *type_counter; /* inc for each call of one type */

int *non_circle_unit; /*order of units without set_mp  */
int *pump_unit; /*unit number Pumps+collecting_pipes */

int *div_unit; /* max. diverters */
int *controller_unit; 

double **out_present;
double **out_past;
double **pointer;    /* for out_n,out_o */
double **deriv_o;  /*DERIVATIVES_OLD */
double **deriv_n;  /*DERIVATIVES_NEW*/
double **init_energy; /* = mc*T_init */
double ***qp_sum;      /* energy flow from each path to node */
double **Qi; /* energy flow */
double **iam;  /* for iam-table */
/* int **point_dead; / *dead_time comp.  point_dead[nr][IN/OUT] */
double **Q;     /* dead_time-energy Q[nr][1..DEAD_TIME_MAX+1] */
double **dummy=NULL; /* dummy array used in pipe2 */
/* moved to global.h
struct therm_hydraulic_struct {
 int unit;
 int type;
 double value1;
 double value2;
 double value_array[5];
};
*/
/* declaration of structs, see global.h or global.orig.h */
struct therm_hydraulic_struct **thermo_struct; /*e.g: thermo_struct[unit].type*/
struct sim_struct *info; /* see global.h */
struct sim_struct **info_array; /* array info sim_struct */
struct radiation_struct *radiation;
struct interface_struct *interface ; /* e.g. interface[unit]->parameter[1] */

/* int ret;  return value */

/*---------------------------------------------------end of declarations---*/

if (argc>1) /*  number of arguments 1:sim 2:option1 ... */
 {
 printf("usage: input file is 'sim.dek',-d: debug flag \n");
 printf("arg1=%s\n",argv[1]); /* pointers to arg1,arg2.. */
 printf("arg2=%s\n",argv[2]); 
 if (strstr(argv[1],"-d")!=NULL) a=1;
 if (argc>2) exit(0);
 }

printf(" usage:\n" /* auto edit by version control */
 " usage: ColSim Version:0.64fluidsystem Date:05/31/02 Time:02:47:24 PM " 
     "\n");
 
 
type           =malloc_1int(UNIT_MAX+2);
unit_order     =malloc_1int(UNIT_MAX+2);
select_unit    =malloc_1int(UNIT_MAX+2);
type_index     =malloc_1int(UNIT_MAX+2);
type_counter   =malloc_1int(UNIT_MAX+2);

non_circle_unit=malloc_1int(UNIT_MAX+2);
pump_unit      =malloc_1int(UNIT_MAX+2);
div_unit       =malloc_1int(UNIT_MAX+2);
controller_unit=malloc_1int(UNIT_MAX+2);


active_circle  =malloc_2int(PUMP_NR+1 ,UNIT_MAX+1);
input_unit     =malloc_2int(UNIT_MAX+1,INPUT_MAX+1);
input_out_nr   =malloc_2int(UNIT_MAX+1,INPUT_MAX+1);
out_label      =malloc_3char(UNIT_MAX+1,INPUT_MAX+1,30);
/*
point_dead     =malloc_2int(DEAD_MAX+1,2);        IN=0,OUT=1 
*/

timezone=malloc_1double(2);
lenght  =malloc_1double(2);
latitude=malloc_1double(2);

parameter      =malloc_2double(UNIT_MAX+1,PAR_MAX+1);
out_present    =malloc_2double(UNIT_MAX+1,OUT_MAX+1);
out_past       =malloc_2double(UNIT_MAX+1,OUT_MAX+1);
deriv_o        =malloc_2double(UNIT_MAX+1,DERIV_MAX+1);
deriv_n        =malloc_2double(UNIT_MAX+1,DERIV_MAX+1); 
init_energy    =malloc_2double(UNIT_MAX+1,DERIV_MAX+1);
Q              =malloc_2double(DEAD_MAX+1,DEAD_TIME_MAX+2);
Qi             =malloc_2double(NODE_MAX+1,10);

qp_sum         =malloc_3double(UNIT_MAX+1,DERIV_MAX+1,PATH_MAX+1);

iam            =malloc_2double(NR_TETA_MAX,NR_PHI_MAX);

/* therm_hydraulic_struct **thermo_struct; 
   allocate just UNIT_MAX pointers, optional in type_init section
   you can allocate space for entries with thermo_struct[unit],
   use it with: e.g. (thermo_struct[unit]).value1=3;
*/
thermo_struct=(struct therm_hydraulic_struct**) 
  malloc(sizeof(struct therm_hydraulic_struct*)*(UNIT_MAX+1) );
for (i=0;i<=UNIT_MAX;i++) thermo_struct[i]=NULL; 

info_array=(struct sim_struct**) 
  malloc(sizeof(struct sim_struct*)*(UNIT_MAX+1) );
for (i=0;i<=UNIT_MAX;i++) {
 info_array[i]=(struct sim_struct*)
  malloc(sizeof(struct sim_struct)*1);
 if (info_array[i]==NULL) { 
   printf("main: malloc error info_array\n"); exit (-1);}
 info_array[i]->unit=0; 
}
radiation=(struct radiation_struct*) 
 malloc(sizeof(struct radiation_struct));
radiation->mode=0; /* init for testing */

interface=(struct interface_struct*)
 malloc(sizeof(struct interface_struct)*(UNIT_MAX+1) );
interface[1].parameter= (double*) parameter[1];
for (i=1;i<=UNIT_MAX;i++) { /* interface struct */
 interface[i].parameter= (double*) parameter[i];
 interface[i].connected_unit=input_unit[i];
 interface[i].out_number_of_connected_unit=input_out_nr[i];
 interface[i].out_current=out_present[i];
 interface[i].out_last_timestep=out_past[i];
 interface[i].state_current=deriv_n[i];
 interface[i].state_last_timestep=deriv_o[i];
 interface[i].init_energy=0;
 interface[i].sys_gains=0;
 interface[i].sys_energy=0;
 interface[i].m_gains=0;
 interface[i].m_flow_internal=0;
} /* all units */

interface[1].parameter[2]=0.0; /* init for testing */
/* interface[1].parameter[1]=0; / *  init for testing 
parameter[1][1]=1;
*/
printf("OK:%.1f\n",interface[1].parameter[2] );
{ /* -----------------------------check array defaults -----------------------*/
if (a) printf("UNIT_MAX:%d OUT_MAX:%d\n",UNIT_MAX,OUT_MAX); 
for (i=0;i<=UNIT_MAX;i++) {
 for (i1=0;i1<=OUT_MAX;i1++) {
  if (   (fabs(out_present[i][i1])>0 ) 
      || (fabs(out_past   [i][i1])>0 )
     )
     { printf("out_n[%d][%d]:%.1f \n",i,i1,out_present[i][i1]);
       printf("out_o[%d][%d]:%.1f \n",i,i1,out_past   [i][i1]);
       exit (-1);
     }
  }
 for (i1=0;i1<=DERIV_MAX;i1++) {
   if (   (fabs(deriv_o[i][i1])    >0)
       || (fabs(deriv_n[i][i1])    >0)
       || (fabs(init_energy[i][i1])>0) 
      )
      {
       printf("deriv_o[%d][%d]:%.1f \n",i,i1,deriv_o[i][i1]);
       printf("deriv_n[%d][%d]:%.1f \n",i,i1,deriv_n[i][i1]);
       printf("init_energy[%d][%d]:%.1f \n",i,i1,init_energy[i][i1]);
       exit (-1);
      }
  }
 }
}


if (a) printf("UNIT_MAX:%d OUT_MAX:%d\n",UNIT_MAX,OUT_MAX); 
type_def_str   =malloc_2char(TYPES_MAX+1,14);
if (type_def_str==NULL) printf("shit\n");

check=init_type_def_str(type_def_str); /* get TYPE discription */

/*-- read dek-File*/
un_max=dek_reader(type,parameter,input_unit,input_out_nr,
                  out_present,out_past,deriv_o,deriv_n,out_label);

if (un_max>= UNIT_MAX)
  {
    printf("exit, to much units in dek , see global.h UNIT_MAX \n");
    exit (-1);
  }

printf("'sim.dek' has %d units and runs with/without win32flag:%d \n"
     ,un_max,WIN32);
/*----------------------------------Read Simulation-Time   -------*/
for (i=1;i<=un_max;i++)   /*SEARCH SIMULATION_CONTROL SIMCONTROL ---*/
  { if (type[i]==62)
     {
     /* xfig.awk etc/xfig.ssc

     TYPE 62

     PAR1: 880401 [yymmdd] date_start (year/month/day) 
     PAR2: 880404 [yymmdd] date_stop (year/month/day)                       
     PAR3: 000000 [] time_start (hour/minute/sec)
     PAR4: 235959 [] time_stop (hour/minute/sec) 
     PAR5: 1000 [] max_error
     PAR6: 1 [] unix_flag
     PAR7: 10 [s] simulation time step
     PAR8: 3600 [s] FastIteration Mode with skip_time_step par8
     PAR9: 0.5  [K] eps_interrupt stop fast iteration
     */


       /* printf("TYPE(%d)\n",type[i]); */
       date_start=(int) parameter[i][1]; /*year/month/day*/
       date_stop =(int) parameter[i][2];
       time_start=(int) parameter[i][3]; /*hour/minute/sec*/
       time_stop =(int) parameter[i][4];
       max_error =      parameter[i][5]; /* error_printing */
       unix_flag =(int) parameter[i][6]; /* 1:unix, 2:os/2 */
       if ( parameter[i][7] != 0 ) h=parameter[i][7]; /*sim time step */
       if ((int) parameter[i][8]==0) FastIteration_bak=FastIteration=0;
       else {
         FastIteration_bak=FastIteration=1;
         skip_time_intervall=parameter[i][8];
       } 
       eps_interrupt=   parameter[i][9]; /* eps fast iteration */
       if ( (FastIteration==1) && (eps_interrupt==0) ) {
         printf("main.c par9 of sim_control error\n"); exit(-1);
       }
       if ( max_error == 0 ) max_error = 1; /* [Ws] */
       printf("\ndate_start: %d\n" "date_stop : %d\n" "time_start: %6d\n"
              "time_stop : %6d\n" "sim-time-step:%.1f\n\n"
              ,date_start ,date_stop ,time_start ,time_stop ,h);
       break;
     };
  } ; /* i=1;i<=un_max;i++ */
if (i>un_max) { printf("no unit 'sim-control' !!\n");exit (-1);}

/*------------------------------------time control---------------*/
/*940101*/
year_start =(int) ( (date_start-1e+6*(int) (date_start/1e+6))/10000);

month_start=(int) ( (date_start-1e+4*(int) (date_start/1e+4)) /100 );
month_stop =(int) ( (date_stop -1e+4*(int) (date_stop /1e+4)) /100 );
if (month_stop<month_start) month_stop+=12;

day_start  =(int) ( (date_start-1e+2*(int) (date_start/1e+2)) /1   );
day_stop   =(int) ( (date_stop -1e+2*(int) (date_stop /1e+2)) /1   );

hour_start =(int) ( (time_start-1e+6*(int) (time_start/1e+6))/10000);
hour_stop  =(int) ( (time_stop -1e+6*(int) (time_stop /1e+6))/10000);
if (   (hour_start>23) 
    || (hour_start<0 ) ) { printf("wrong time_start\n");exit(-1);}
if (a) printf("hour_start:%d\n",hour_start);

min_start  =(int) ( (time_start-1e+4*(int) (time_start/1e+4)) /100 );
min_stop   =(int) ( (time_stop -1e+4*(int) (time_stop /1e+4)) /100 );

sec_start  =(int) ( (time_start-1e+2*(int) (time_start/1e+2)) /1   );
sec_stop   =(int) ( (time_stop -1e+2*(int) (time_stop /1e+2)) /1   );

/* search day_of_year */
i_month=1;
while (i_month<month_start)
  {
     day_of_year+= no_days_of_month[i_month];
    i_month++;
  }
/*-----------------------------------------------------------------*/
if (a) printf("main:%d i_month:%d month_start:%d\n",day_of_year,i_month,month_start);

nr_of_unit_calls=unit_order_init(0,parameter,input_unit,input_out_nr,
           out_present,out_past,deriv_o,deriv_n,h,init_energy,
           qp_sum,tsim,init_flag,type_def_str,type,un_max,unit_order,
           &first_model_index,&last_model_index  );

if (a) printf("main: unit_order_init done nr_of_unit_calls:%d\n"
              ,nr_of_unit_calls);
if (a) fflush(stdout);

/*-------------------------------------end_of init----------------------------*/


/* hydraulic Init  
SetUpHydraulicEquations(type,unit_order,parameter,out_present,nr_of_unit_calls);
exit (-1);
*/


if (1) /* Get type_index: */
 {
 int unit;
 /* laufunit=unit_order[i]; */
 for (i=1;i<UNIT_MAX;i++) type_index[i]=type_counter[i]=select_unit[i]=0; 

 for (i=1;i<=nr_of_unit_calls;i++) 
  {
  unit=unit_order[i];
  select_unit[unit]++;
  if (select_unit[unit]==1) type_counter[ type[unit] ]++; /*  */
  if (select_unit[unit]==1) type_index[unit]=type_counter[  type[unit] ];
  if (a) printf("index:%3d \tunit:%3d \ttype:%3d \ttype_index:%3d \t%s\n",i
         ,unit,type[unit],type_index[unit],type_def_str[type[unit]]);
  }
 if (a) printf("first_model_index:%d \n",first_model_index);
 if (a) printf("last_model_index:%d \n",last_model_index);
 }


printf("\n --------------- START SIMULATION-------month_start:%d month_stop:%d  \n"
         ,month_start,month_stop);
if (a) fflush(stdout);


/*-----simulation loop , timesteps -------------------------------------------*/

for (i_month=month_start;i_month<=month_stop;i_month++)
  {
  if (i_month>12) j_month=i_month-12;
  else j_month=i_month;

  if (i_month<month_stop) day_stop2=no_days_of_month[i_month];
  else                    day_stop2=day_stop;

  if (i_month==month_start) day_start2=day_start;
  else                      day_start2=1;

  day_of_year+=day_start2-1;
  for (i_day=day_start2;i_day<=day_stop2;i_day++)
    {
    day_of_year++;
    if (a) printf("day_of_year:%d\t",day_of_year);

    day_of_week=  day_of_year -7*( (int) (1+(day_of_year/7)) -1 )
                 + first_day_of_year;
    if (day_of_week > 7) day_of_week-=7;

    if  (day_of_week==7) week_flag=1;
    else                 week_flag=0;

    /*-----------------print date/time ---------------*/
    if (a) { for (i=0;i<=8;i++)
                 printf("%c",month_string[i_month][i]);
             printf("; ");
             for (i=0;i<=8;i++) printf("%c",day_string[day_of_week][i]);
             printf(",the ");
             printf("%d/%d/%d \n", i_day,i_month,year_start);
           }


    if (  (i_month==month_start) &&(i_day  ==  day_start))
       t_start=(double) (hour_start*3600+min_start*60+sec_start);
    else
       t_start=0;

    if (  (i_month==month_stop) &&(i_day  ==  day_stop))
       t_stop=(double) (hour_stop*3600+min_stop*60+sec_stop);
    else
       t_stop=(double) (24*3600-1);


/*-----daily loop from tsim=1sec to 24*3600sec with timestep h ---------------*/
    for (tsim=t_start;tsim<=t_stop;tsim+=h)
      {
            
      i_hour=(int)   (tsim/3600);
      i_min =(int) ( (tsim-3600* (double) (i_hour) )/60);
      i_sec =(int) (tsim-3600*(double) (i_hour)-60*(double) (i_min)) ;

      if (tsim==0) day_flag=1;
      else         day_flag=0;

      /* set hour-change and 10_minutes_intervalls */
      if( (!i_min) && (!i_sec) ) hour_flag=1;
      else                       hour_flag=0;

      if ( (!(i_min-10*(int) (i_min/10))) && (!i_sec) ) min10_flag=1;
      else                                              min10_flag=0;

      stop_flag=((tsim>t_stop-h)&&(i_month==month_stop)&&(i_day==day_stop2));

   /*-----------------------------------------------------case type-----*/

   /* select iteration algorithm */
/* !!!!!!!!!!!!!  test  
if   (tsim>=60) { FastIteration=1; NoOfIterations=300; }
else            { FastIteration=0; NoOfIterations=1; }
if   (tsim>=80) skip_flag=1;
*/

      /* init internal iteration for this timestep tsim */
      if (tsim<=t_start+h) FastIteration=0;  /* !!! */
      else         FastIteration=FastIteration_bak;
      i_iteration=1; 
      stop_iteration=0;
      skip_timer-=h; /* count down */
      if (skip_timer<=0) skip_flag=0;
      else               skip_flag=1; /* no iteration */
      if (a) printf("tsim:%.1f skip_timer:%.1f skip_flag:%d\n"
            ,tsim,skip_timer,skip_flag);
      /* ------------------------------------ if Loop Iteration --------------*/
      do { /* while (!stop_iteration) --*/
         if (a) printf("\nmain.c:######### invoking iteration,"
                       "#######################  tsim:%.1f [SecOfDay]\n" ,tsim);
         if (FastIteration) { /* select unit_order start&end 
               FastIteration: select discrete timestep and iterate
               inside up to convergence, skip following timesteps
               for models but not for printers
             */
            if ( i_iteration>NoOfIterations) {
             printf("MaxNumber of Iterations NoOfIterations:%d exceeded\n",
                     i_iteration); exit (-1);
            }
            if (skip_flag) { /* just call printers */
             unit_order_start=last_model_index+1;
             unit_order_stop=nr_of_unit_calls;
             stop_iteration=1; /* exit FastIteration loop */
             }
            if (!skip_flag) { /* get eps from deriv's */
              /* get current error */
              eps=convergence(deriv_o,deriv_n,out_past,out_present,
                              eps_interrupt,i_iteration,
                              unit_order,first_model_index,last_model_index,a); 
               /* exit-criteria, but check even number of iterations
                 to ensure correct out_n-pointer and take minimum 2 iter. */
              if ( (eps<eps_interrupt)&&(i_iteration%2==1)&&(i_iteration>1) ) {
               stop_iteration=1;                     /* eps reached */
               skip_flag=1;                /* skip to next iteration timestep */
               skip_timer=skip_time_intervall; /* count down set */
               }
              if (i_iteration==1) unit_order_start=1; /* call initial-types */
              else                unit_order_start=first_model_index; /* run */
                                                              /* thru models */
              if (!stop_iteration) unit_order_stop=last_model_index;
              else                 unit_order_stop=nr_of_unit_calls;
              if (a) printf("tsim:%.1f ##eps:%.4f eps_interrupt:%.4f "
                            "stop_iter:%d i_iter:%d \n"
                     ,tsim,eps,eps_interrupt,stop_iteration,i_iteration);
            } /* if (!skip_flag) */

         } /* if FastIteration */
       else /* !FastIteration */ {
         if (a) printf("no FastIteration\n");
         unit_order_start=1;
         unit_order_stop=nr_of_unit_calls;
         stop_iteration=1; /* just one call of models per timestep */
       }


      /* reset select  */
      for (i=1;i<=nr_of_unit_calls;i++) select_unit[i]=info_array[i]->select_unit=0; 
      if (a) printf("tsim:%.1f i_iteration:%d NoOfIterations:%d Tcoll_o:%.4f "
                    "Tcoll_n:%.4f \n" ,tsim,i_iteration,NoOfIterations,
                    deriv_o[8][1],deriv_n[8][1]);


      for (i=unit_order_start;i<=unit_order_stop;i++)
      /* for (i=1;i<=nr_of_unit_calls;i++) / * nr_of_unit_calls  
         go through units of dek */
        {
        laufunit=unit_order[i];
       info_array[laufunit]->day_of_year=day_of_year;
        if (a) printf("unit_order_index:%23d unit:%3d select_unit:%1d %s\n" 
                      ,i,laufunit,select_unit[laufunit]
                      ,type_def_str[type[laufunit]]);
        if (a) fflush(stdout);
/*
for (i=1;i<=nr_of_unit_calls;i++) printf("unit_order_index:%2d current unit:%d  \n" ,i,unit_order[i]);
if (a) fflush(stdout);
*/
        info_array[laufunit]->init_flag=0;
        if (init_flag) { /* set info_array */
         int i;
         info_array[laufunit]->init_flag=1;
         info_array[laufunit]->unit=laufunit;
         info_array[laufunit]->type=type[laufunit];
         for (i=0;i<12;i++) info_array[laufunit]->info[i]=type_def_str[type[laufunit]][i];
         info_array[laufunit]->info[11]='\0';
         info_array[laufunit]->h=h;
         info_array[laufunit]->previous_unit=previous_unit;
    
#if 0
 int unit; /* unique ID for the objects */
 int type; /* model type */
 char info[12]; /* model info string */
 int init_flag; /* =1, if sim-start (tsim=tstart) */
 double h; /* sim time step */
 double tsim; /* [s] sim time of the day */
 int day_of_year;
 double previous_unit; /* this unit was processed just before */
 int active_plug_input; /* current plug is pushed to this input */
 int select_unit; /* number of call of this unit due to this time step */
 int type_index; /* index for same types, internal array index */
 int last_time_flag; /* last call of this unit due to this timestep */
#endif

        }
        
        info=(struct sim_struct*) info_array[laufunit]; /* point to array */
        info->tsim=tsim;
        type_call(
              info,
              interface,
              radiation,
              type_def_str,
              laufunit, /* unit */
              un_max,
              type,
              select_unit,
              type_index,
              init_flag,
              timezone,
              lenght,
              latitude,
              day_of_year,
              year_start,
              j_month, /* int month */
              i_day,   /* day */
              i_hour,  /* hour*/
              i_min,   /* min */
              i_sec,   /* sec */
              day_flag,
              hour_flag,
              week_flag,
              day_of_week,
              min10_flag,
              stop_flag,
              previous_unit,
              input_unit, /* int **in_u */
              input_out_nr, /*int **in_nr */
              iam,
              h,
              tsim,
              parameter, /* double **par */
              out_present, /* double **out_n */ 
              out_past,  /* double **out_o */
              deriv_o,
              deriv_n,
              init_energy,
              Qi,
              Q, /* pipe */
              qp_sum,
              a,
              unit_order,
              nr_of_unit_calls,
              dummy,
              out_label,
              skip_time_intervall,
              thermo_struct
             ) ;
         previous_unit=laufunit;
        } /*next lauf_unit*/

     mp_change=0; /*reset mp-changes caused by controller */

     if (init_flag)
       {
         Ausgabe_Input(input_unit,input_out_nr,un_max,type,type_def_str);
         /*
          Ausgabe_Parameter(parameter,un_max,type,type_def_str);
         */
         printf("\n Output-difference after Initializing \n");
         Ausgabe_Out_pr(out_present,out_past,un_max,type,type_def_str,0);
         /*
          print_derivs(deriv_o,un_max,type,type_def_str);
          print_inp_const(out_present,out_past,un_max,type,type_def_str);
         */
       };


     if (a==1) /* check outs NaN */
      {
      int i3,i4;
      for (i3=1;i3<=un_max;i3++)
       for (i4=1;i4<=OUT_MAX;i4++)
        if ( isnan(out_present[i3][i4] )!=0) 
          { printf("main: tsim:%.1f NaN murx unit:%d out_nr:%d\n"
             ,tsim,i3,i4); exit (-1); }
        if ( isinf(out_present[i3][i4] )!=0) 
          { printf("main:Inf murx unit:%d out_nr:%d\n",i3,i4); exit (-1); }
      
      }
      /* energy balance over all types */
      if (!FastIteration_bak) {
dQ_error=energy_outcome(laufunit,out_present,tsim,
                             init_flag,stop_flag,
                             h,un_max,init_energy,out_past,type,type_def_str,
                             max_error,qp_sum,deriv_o,deriv_n,input_unit,
                             input_out_nr);
}

      pointer=out_present; /* switch outs  */
      out_present=out_past;
      out_past=pointer;

      pointer=deriv_n;     /* switch derivs */
      deriv_n=deriv_o;
      deriv_o=pointer;

      init_flag=0;
      i_iteration++;
      } while (!stop_iteration); 

      } /* end tsim=tstop */
    }; /*end i_day */
  }; /*end i_month */

 /*--------------------------------stop_sim----------------------------*/

   Ausgabe_Out_pr(out_present,out_past,un_max,type,type_def_str,0);

   printf("tsim=%f \n",tsim);
   printf("System-Energy-Check: %.4e [Ws]\n",dQ_error);

/* several OS or compilers doesn't like those things:
free(parameter);free(input_unit);free(input_out_nr);
free(out_present);free(out_past);free(weather);
free(load);
*/ 

return 0;
} /* end main colsim */



double convergence(double **deriv_o,double **deriv_n,
                   double **out_o,double **out_n,
                   double eps_interrupt,
                   int i_iteration,int *unit_order,
                   int first_model_index,int last_model_index,int a)
{
/*
   convergence tries to check change of deriv_[][1..n] for all active units.
   if 'second_call_of_this' is set, deriv_max[unit_order] is created
   thru non-zero deriv's of each unit. in the following deriv_max is
   used to search for error eps_max=max(eps).
*/
double eps_set=eps_interrupt; /* 0.01 : T_o=10.011 T_n=10.020 is OK */ 
double eps;
double eps_max=0; /* maximum error of all deriv's */
int i,i1;
int unit;
static int local_init_flag=1;
static int second_call_of_this=0;
static int deriv_max[UNIT_MAX]; /* check number of diff-equ's */

/*  a=0; */

if (local_init_flag) { /* init */
 for (i=0;i<UNIT_MAX;i++) deriv_max[i]=0; /* reset array */
}
if (second_call_of_this==1) { /* check state of deriv's */
 for (i=first_model_index;i<=last_model_index;i++) { /* i thru unit_order */
  unit=unit_order[i];
  for (i1=1;i1<DIV_NR;i1++) { /* run thru deriv's */
   if (deriv_n[unit][i1]!=0) deriv_max[i]++;
  }
  if (a) printf("convergence: init unit:%d deriv_max:%d T:%.4f \n"
                ,unit,deriv_max[i],deriv_n[unit][deriv_max[i]]);
 }
 second_call_of_this=0;
} /* second_call_of_this */

if (a) printf("convergence:%d\n",i_iteration);
if (i_iteration>100) exit(-1);
/* check transient state of marked deriv's */
if (i_iteration==2) { /* rewrite old stat of controllers equations etc */
 for (i=1;i<first_model_index;i++) { 
  int out_nr=1;
  unit=unit_order[i];
  for (out_nr=1;out_nr<OUT_MAX0;out_nr++) { /*scan outputs */
   if (out_n[unit][out_nr]!=out_o[unit][out_nr]) {
    printf("convergence: flipping: unit:%d out:%d value:%.4f\n",
           unit,out_nr,out_n[unit][out_nr]);
    out_n[unit][out_nr]=out_o[unit][out_nr]; /* this is the 2nd call !!! */
    printf("%d,%d; %.4f=%.4f\n",unit,out_nr
           ,out_o[unit][out_nr],out_n[unit][out_nr]);
   }
  }
 } /* i=1;i<first_model_index */
} /* i_iteration==2 */
for (i=first_model_index;i<=last_model_index;i++) { /* i thru unit_order */
  unit=unit_order[i];
  if (deriv_max[i]>0) { /* check deriv existance */
   if (a) printf("unit:%d T1..deriv_max: ",unit);
   for (i1=1;i1<=deriv_max[i];i1++) { /* run thru deriv's */
    if (a) printf("%.4f ",deriv_n[unit][i1]);
    eps=fabs(deriv_n[unit][i1]-deriv_o[unit][i1]);
    if ( (a) && (eps<eps_set) ) printf("OK, ");
    if (eps>eps_max) eps_max=eps;
   }
  if (a) printf("\n");
  }
}

if (local_init_flag) {  /* set second_call_of_this */
 local_init_flag=0; 
 second_call_of_this=1; 
 eps_max=10; /* first run */
 }
return (eps_max);
} /* convergence() */
