/* diverter.c */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/

#include "global.h" 
#include <stdio.h>
#include <stdlib.h>

/*
011108 flow_div converted to diverter.c
020109 labels update, macros update , hydraulic subfunction
020131 comment update
020325 exit on mode!=1
020517 if (contr>1) contr=1; if (contr<0) contr=0;

*/
int diverter_hydraulic(int unit,double **par,double position,double *Ceta_out0
                       ,double *Ceta_out1);

/*       10        20        30        40        50        60        70       */

/*--------------------------------------FLOW_DIVERTER--TYPE 11 --------*/
int diverter(unit,par,in_u,in_nr,out_o,out_n,init_flag)
           int    unit;      /* current unit number */
           double **par;     /* parameter array par[unit][1..PARMAX]  */
           int    **in_u;    /* input unit array, input connections   */
           int    **in_nr;   /* input number array                    */
           double **out_o;   /* output array with privious state      */
           double **out_n;   /* output array with current state       */
           int    init_flag; /* set at initial loop of sim            */

{
/*  instuctions for xfig-object, updated with xfig.awk
  
    TYPE 11
    PRIORITY  6          used by config routine
    HYDRAULIC 11         diverter fluid 
  
    PAR1: 1    [1] 'mode' mode 1: T_piece
    PAR2: 0    [1] 'nop' not used
    PAR3: 1.0e+07 [1/m^4] 'CetaA2'
    PAR4: 1.0  [] 'a1' quadr_coeff mp_0
    PAR5: 0.0  [] 'a2' linear_coeff mp_0
    PAR6: 0.001 [] 'a3' offset_coeff mp_0
    PAR7: -1.0  [] 'b1' quadr_coeff mp_1
    PAR8: 0.0  [] 'b2' linear_coeff mp_1
    PAR9: 1.0  [] 'b3' offset_coeff mp_1    
    PAR10: 0 [1]  'nop' 
    PAR11: 1 [1]  'dat1' Diverter[i].dat port0
    PAR12: 2 [1]  'dat2' Diverter[i].dat port1

    INP1: 0 [Ws] 'dQin' deltaQ-inlet
    INP2: 0 [kg/s] 'mdot_in' mp-inlet
    INP3: 0.5 [] 'ctr' 0:mp0_out =mp 1:mp1_out 
    INP4: 0.5 [] 'pos' 0:mp0_out =mp 1:mp1_out 

    OUT1: [Ws] 'dQ0out' deltaQ0-outlet
    OUT2: [kg/s] 'mdot0out' mp0_out =mp, if contr=0
    OUT3: [Ws] 'dQ1out' deltaQ1-outlet
    OUT4: [kg/s] 'mdot1out' mp1_out =mp, if contr=1
    OUT5: [kg/h] 'mdotout_view' 

    OUT6: [Ws] 'sysgain' system inlet energy 
    OUT7: [Ws] 'sysloss' system outlet energy 
    OUT8: [Ws] 'syseneg' internal energy 
    OUT9: [1] 'Ceta0' f(ctr)
    OUT10: [2] 'Ceta1' f(ctr)
    OUT11: [2] 'deltaP' written by hydraulic type
    
*/

/*----------------------------------------------------------------------------*/
int a=0; /* debug */
int mode=(int) par[unit][1];

double Qin     =out_n IN(1);
double mdot_in =out_n IN(2);
double contr   =out_n IN(3);
double position=out_n IN(4);

double Qout0,mdot0_out,Qout1,mdot1_out;
double Ceta_out0=0,Ceta_out1=0;

int hydraulic_calc=(in_u[unit][4]==unit); /* is inp 4 connectod ? */


if (init_flag) {
  out_n[unit][9]=3.3;out_n[unit][10]=2.2;
  if (mode!=1) { 
   printf("diverter[%d] unknown mode:%d \n",unit,mode); exit (-1);
  }
}

if (hydraulic_calc) 
 diverter_hydraulic(unit,par,position,&Ceta_out0,&Ceta_out1);

if (a) printf("diverter[%d]\n",unit);
if (contr>1) contr=1;
if (contr<0) contr=0;

Qout0=out_n IN(1)*(1-contr);
mdot0_out=(1-contr)*mdot_in;
Qout1=Qin-Qout0;
mdot1_out=mdot_in-mdot0_out;

/* set ouputs */
out_n[unit][1]=Qout0;
out_n[unit][2]=mdot0_out;
out_n[unit][3]=Qout1;
out_n[unit][4]=mdot1_out;
out_n[unit][5]=mdot0_out*3600; /* kg/h */


out_n[unit][9]=Ceta_out0;
out_n[unit][10]=Ceta_out1;

if (   (mdot_in<0) || (mdot0_out<0) || (mdot1_out<0)) {
 printf("diverter[%d]: mdot neg: mdot_in:%.4fkg/h mdot0_out:%.4f "
        "mdot1_out:%.4f ctr:%e\n",
        unit,mdot_in*3600,mdot0_out*3600,mdot1_out*3600,contr);
 exit (-1); 
}
return 1;
}
/*----------------------------------------------------------------------------*/

int diverter_hydraulic(unit,par,position,Ceta_out0,Ceta_out1)
           int    unit;      /* current unit number */
           double **par;     /* parameter array par[unit][1..PARMAX]  */
           double position;
           double *Ceta_out0;
           double *Ceta_out1;
{
int a=0;
double cetaA2,a1,a2,a3,b1,b2,b3,help1,help2;
if (a)  printf("unit :%d  position=%.4f\n", unit,position);

cetaA2=par[unit][3];
a1=par[unit][4];
a2=par[unit][5];
a3=par[unit][6];
b1=par[unit][7];
b2=par[unit][8];
b3=par[unit][9];

/*Ceta_0 dann Ceta_1  position=1 heisst mp1=mp_in*/
help2=b1*position*position+b2*position+b3;                
/*position=1 ceta_mp1=infinity position=0  mp0=mp_in    */
*Ceta_out0=cetaA2/((help2*help2)+1.0e-7); 

help1=(a1*position*position+a2*position+a3);
     /*position=0 ceta_mp0=infinity position=1  mp1=mp_in    */
*Ceta_out1=cetaA2/((help1*help1)+1.0e-7);

if (a) printf("\n\nflow_div.c:UNIT %d position:%.4f Ceta_out0:%6.3f "
              "Ceta_out1: %6.3f "
              ,unit,position,*Ceta_out0,*Ceta_out1);
return 1;
}

