/* heatexchanger.c */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/

#include "global.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
001103 chriwi bug with firsttimeflag , select_unit integrated 

*/
/*----------------------HEAT_EXCH TYPE 10 ----STATIC HEATEXCHANGER--------*/
int heatexchanger(unit,par,in_u,in_nr,out_n,out_o,deriv_o,deriv_n,h,
		  init_energy,qp_sum,tsim,init_flag,previous_unit,select_unit)
           int    unit;      /* current unit number */
           double **par;     /* parameter array par[unit][1..PARMAX]  */
           int    **in_u;    /* input unit array, input connections   */
           int    **in_nr;   /* input number array                    */
           double **out_o;   /* output array with privious state      */
           double **out_n;   /* output array with current state       */
           double **deriv_o; /* derivatives: used for diffential equ. */
           double **deriv_n; /*  new state of deriv                   */
           double h;         /* sim time step                         */
           double **init_energy; /* initial energy in RC-comp. [Ws]   */
           double ***qp_sum; /* energy to one node qp[unit][n][m]     */
           double tsim;      /* simulation time in [s] of this day    */
           int    init_flag; /* set at initial loop of sim            */
           int previous_unit; /* get unit which is connected with inp i */
           int    *select_unit; /* number of calls in this timestep   */

{
/*  ./etc/xfig.awk         
    
   TYPE 10
   PRIORITY  6          used by config routine
   HYDRAULIC 10         4 pole fluid            

    
    PAR1: 1 [] mode 1=> counterflow
    PAR2: 4.19 [kJ/kgK] cpHot
    PAR3: 4.19 [kJ/kgK] cpCold
    PAR4: 300 [W/K] kA heat transfer coeff. area product
 
    INP1: 0 [Ws] deltaQ0-inlet hot 
    INP2: 0 [kg/s] mp0-inlet hot
    INP3: 0 [Ws] deltaQ1-inlet cold
    INP4: 0 [kg/s] mp1-inlet cold
    INP5: 10 [C] temp ambient

    OUT1: [Ws] deltaQ-outlet0 hot
    OUT2: [kg/s] mp-outlet0 hot
    OUT3: [Ws] deltaQ-outlet1 cold
    OUT4: [kg/s] mp-outlet1 cold
    OUT5: [] effectiveness 
    OUT6: [Ws] system inlet energy 
    OUT7: [Ws] system outlet energy 
    OUT8: [Ws] internal energy
    OUT9: [W] P_exchg + loss
    OUT10: [C] Temp Hot Out
    OUT11: [C] Temp Cold Out
    OUT15: [W] Pexchange
    OUT16: [W] Ploss
    OUT17: [C] deltaT  
    OUT18: [] noHeatTransferFlag internal
    OUT19: [s] first_execute_time
    OUT20: [C] Temp Hot In
    OUT21: [C] Temp Cold In

    DERIV1: 10 [C] temp of hot in
    DERIV2: 10 [C] temp of cold in
*/
/*
 function: first heatexchanger-called--> return (nothing done)
           second time call: (next pump circle) calculation
 bugs:
 if   mp1=50& mp2=0
 then mp1=0 & mp2=50 at same time --> error energy bilance 
 kA_loss only for hot side !!, no heat transfer , if mp==0
 eff always 11
 Cmin,Cmax,NTU ??
*/ 

int a=0; /*-----------------a=1 for analyse */
double cpHot,cpCold,kA,dQinHot,mpinHot,dQinCold,mpinCold;
double  dQoutCold=0,dQoutHot=0;
double TempHotOut=10,TempColdOut=10;
double Q11=0,Q21=0,Q13=0,Q23=0;
double Q_transfer=0;
double TempHotIn=10,TempColdIn=10,Cmax,Cmin,ratio,NTU,eff=0;

int noHeatTransferFlag; /* if mpchange : mp1==0 mp2!=0 */


select_unit[unit]++;
if (select_unit[unit]==1)  return(1); /* if first_time_flag=1; */
if (select_unit[unit]==3) return(1);  /*return third loop !! */ 


cpHot =par[unit][2]*1e+3;
cpCold=par[unit][3]*1e+3;

kA=par[unit][4];
  
mpinHot =out_n[  in_u[unit][2] ]
              [ in_nr[unit][2] ];
mpinCold=out_n[  in_u[unit][4] ]
              [ in_nr[unit][4] ];

/* --------------------------------------initialize-------------*/

  if (init_flag)
    { 
      if ( (int) par[unit][1]!=1)
        { printf("HEAT_EXCH wrong mode unit%d\n",unit);exit (-1); }
      deriv_n[unit][1]
     =deriv_o[unit][1]
     =deriv_o[unit][2]
     =deriv_o[unit][2]=out_n[  in_u[unit][5] ]
                            [ in_nr[unit][5] ]; /* set to Tamb */
	
    }

  /* ------------------------------------calculate ----------------*/
  
  dQinHot = 
       out_n[  in_u[unit][1] ]       /* Q_in */
            [ in_nr[unit][1] ];
  dQinCold=     
        out_n[  in_u[unit][3] ]       /* Q_cold */
             [ in_nr[unit][3] ];
  
  /* no heat transfer if one mp == 0 */
  if ( (mpinHot==0) || (mpinCold==0) )  noHeatTransferFlag=1;
  else                                  noHeatTransferFlag=0;
  
     
  if (a) printf("mpinHot:%.2f mpinCold:%.2f dQinHot:%.2f dQinCold:%.2f \n"
                ,mpinHot*3600,mpinCold*3600,dQinHot,dQinCold);
  if (mpinHot>0)  TempHotIn =dQinHot /(mpinHot *h*cpHot );
  if (mpinCold>0) TempColdIn=dQinCold/(mpinCold*h*cpCold);
		
  /* ---------------calculate heat exchanger counter flow------------*/
  /* calculate HeatTransfer mode 1*/ 
  if ( !noHeatTransferFlag  )
   {
    /* a=1; */ 
    if (a) printf("calculate heat exchanger:TempHotIn:%.2f TempColdIn:%.2f "
                  "dQinHot:%.2f mpinHot:%.2f \nderiv_o1:%.2f  "
                  ,TempHotIn,TempColdIn,dQinHot,mpinHot*3600,deriv_o[unit][1]);
		  
    if ( (mpinHot*cpHot) > (mpinCold*cpCold) )
     {
      Cmax=mpinHot *cpHot;
      Cmin=mpinCold*cpCold;
     }
    else
     {
      Cmax=mpinCold*cpCold;
      Cmin=mpinHot *cpHot;
     }
    if (Cmax==0) { printf("Cmax=0: HeatTransfer &&first_time_flag "
                   "mpinHot:%.2f mpinCold:%.2f tsim:%.4f \n"
		   ,mpinHot,mpinCold,tsim);exit(-1); }

    ratio=Cmin/Cmax;   /* C1/C2= R1  bug 6/98 */  
    if (fabs(Cmin-Cmax)<1e-5) ratio=1;
    NTU=kA/Cmin; 
    
    if (ratio==1) eff=NTU/(1+NTU);
    else          eff=( 1-exp( -NTU*(1-ratio) ))/(1-ratio*exp(-NTU*(1-ratio)));

    if ( (eff<0) || (eff>1) ) 
     { printf("eff:%.2f\n"
              "Cmin=%.2f Cmax=%.2f NTU=%.2f tsim:%.1f\n"
	      ,eff,Cmin,Cmax,NTU,tsim); exit (-1); }
    Q_transfer=h*eff*Cmin*(TempHotIn-TempColdIn);

    if (a) printf("eff=%.2f  ------------------------------Q_transfer:%.2f \n"
                  ,eff,Q_transfer);
		  
    dQoutHot =dQinHot -Q_transfer;
    dQoutCold=dQinCold+Q_transfer;

    
    /* TRNSYS:
      TempHotOut=TempHotIn-eff*(Cmin/cpHot)*(TempHotIn-TempColdIn);
       dQoutHot=mpinHot*cpHot*h*TempHotOut;
      TempColdOut=eff*(Cmin/cpCold)*(TempHotIn-TempColdIn)*TempColdIn;
      dQoutCold=dQinHot+dQinCold-dQoutHot;
      TempColdOut=dQoutCold/(mpinCold*cpCold*h);
      but some problems occured for energy bilance
      we try to loose no energy
    */
    } /* end mp != 0 */
  else /* noHeatTransferFlag */  
   {
    dQoutHot =dQinHot;
    dQoutCold=dQinCold;  /* no change */
    
   }
  if (mpinHot>0)  TempHotOut =dQoutHot/(mpinHot*cpHot*h);
  if (mpinCold>0) TempColdOut=dQoutCold/(mpinCold*cpCold*h);
   
  if (a)
   printf("heatexch: TempHotOut:%.2f  TempColdOut:%.2f tsim:%.4f \n"
          "          TempHotIn:%.2f   TempColdIn:%.2f \n"
          "          dQinHot:%.2f     dQinCold:%.2f    \n"
          "          dQinHotHe:%.2f   dQinColdHe:%.2f  \n"
          "          dQoutHot:%.2f    dQoutCold:%.2f   \n"
	  "          mpHotOut:%.2f    mpColdOut:%.2f   \n"
	  "          Q13:%.2f         Q23:%.2f         \n"
                  ,TempHotOut,TempColdOut,TempHotIn,TempColdIn,tsim/3600
		  ,Q11,Q21
		  ,dQinHot,dQinCold
		  ,dQoutHot,dQoutCold
		  ,out_n[  in_u[unit][2] ]
		        [ in_nr[unit][2] ]*3600,out_n[  in_u[unit][4] ]
			                             [ in_nr[unit][4] ]*3600
		  ,Q13,Q23);
		  
  if ( (TempHotOut>150) || (TempColdOut>150) || (TempHotOut<-10) || (TempColdOut<-10) )
     {
     printf("hexchg:%d TempHotOut:%.1f TempColdOut:%.1f HeatTransferFlag:%d\n"
            "TempHotIn:%.1f TempColdIn:%.1f \n"
            ,unit,TempHotOut,TempColdOut,!noHeatTransferFlag,TempHotIn,TempColdIn);
     exit (-1);
     }

  /* --------------------------------------set output -------------*/
  out_n[unit][1]=dQoutHot; /*mpinHot*cpHot*h*TempHotOut; */
   
  out_n[unit][2]= out_n[  in_u[unit][2] ] 
  		               [ in_nr[unit][2] ];
		       
  out_n[unit][3]=dQoutCold; 
  out_n[unit][4]= out_n[  in_u[unit][4] ]
                       [ in_nr[unit][4] ]; 
  out_n[unit][5]=eff;
  /*   out_n[unit][6,7,8]=0;  */
   
  /* Q_transfer (P_trans) */
  if ( !noHeatTransferFlag  ) out_n[unit][9]=Q_transfer/h; 
  else                        out_n[unit][9]=0;

  /* P_echg + loss  */
  out_n[unit][9]= (out_n[  in_u[unit][3] ]
                        [ in_nr[unit][3] ]-dQoutCold)/h;
  if ( fabs( (Q_transfer/h)+out_n[unit][9] >1)  ) 
     printf("Q_transfer:%.1f QWW: %.1f \n",Q_transfer,out_n[unit][9]);

  out_n[unit][10]=TempHotOut; 
  out_n[unit][11]=TempColdOut;
  out_n[unit][15]=Q_transfer/h; /* Pexchange */
  out_n[unit][16]=0;            /* Ploss, static model */
  out_n[unit][17]=(TempHotOut+TempHotIn)/2-(TempColdOut+TempColdIn)/2;  /* deltaT */
  out_n[unit][18]=out_o[unit][18]=noHeatTransferFlag;
  /* first_execute_time in out_n[unit][19] */
  out_n[unit][20]=TempHotIn;
  out_n[unit][21]=TempColdIn;

  return 1;
}


