/* pump.dynamic.c */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/

#include "global.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* 
000515 default mass set to one plug mp_max*h
000528 minimum one plug capacity
000609 bugfix h=1, first_time_execute substituded through select_unit[]
000621 fabs(Tnew-deriv_o[unit][1])>10 from5K
001221 unit input4 kg/h not kg/s
000213 internal energy balance included , exit on mp>mp_max, Qout 2nd timestep
000303 if (mp_new*3600>mp_max*1.1) modified
010517 80 columns limitation
010627 type_index & thermo_struct added
       (*thermo_struct[unit]).cp=Get_cp_from_Temp(unit,METHOD_WATER,20); 
020408 inp hyd added
*/


/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*--pump type 3------------------------------PUMP TYPE 3----------------------*/

int pump(unit,par,in_u,in_nr,out_o,out_n,deriv_o,deriv_n,h,init_energy,qp_sum,
         tsim,init_flag,previous_unit,select_unit,type_index,thermo_struct)
           int    unit;      /* current unit number */
           double **par;     /* parameter array par[unit][1..PARMAX]  */
           int    **in_u;    /* input unit array, input connections   */
           int    **in_nr;   /* input number array                    */
           double **out_o;   /* output array with privious state      */
           double **out_n;   /* output array with current state       */
           double **deriv_o; /* derivatives: used for diffential equ. */
           double **deriv_n; /*  new state of deriv                   */
           double h;         /* sim time step                         */
           double **init_energy; /* initial energy in RC-comp. [Ws]   */
           double ***qp_sum; /* energy to one node qp[unit][n][m]     */
           double tsim;      /* simulation time in [s] of this day    */
           int    init_flag; /* set at initial loop of sim            */
           int    previous_unit; /* coming from this unit             */
           int    *select_unit; /* number of calls in this timestep   */
           int    *type_index; /* index of calling this type          */
           struct therm_hydraulic_struct **thermo_struct;
{
/*  
    TYPE 03
    PRIORITY  5          used by config routine: pump priority
    HYDRAULIC 3          pump fluid


    PAR1: 300 [kg/h] mp_max
    PAR2: 4.19 [kJ/kgK] cp
    PAR3: 360 [W] Pmax_el Power consumption at mp_max
    PAR2: [1] exponent a: Pel=(mp/mp_max)**a*Pmax, mp with [kg/h] 
    PAR5: 0 [1] mode0: ctr=inp3 1:mp_in=inp4[kg/h]/hydraulic,2:auto-select 
    PAR6: 
    PAR7: 5 [kg] mass due to capacity , if 0: auto-default
    PAR8: 1 [1] a, hydraulic-field deltaP=a+b*Vp+c*vp^2
    PAR9: 1 [1] b, hydraulic
    PAR10: 1 [1] c, if a=b=c=0:Vp=fix from inp3 or 4
    PAR11: 1 [-] Pump[i].dat, extern used

    INP1: 0 [Ws] deltaQ-inlet
    INP2: 0 [kg/s] mp-inlet
    INP3: 0 [] controller signal for mode0
    INP4: 0 [kg/h] target mp for mode1
    INP5: 20.0 [C] ambient temperature for starting pump
    INP6: 1 [-] PumpStage (only used for hydraulic)


    OUT1: [Ws] deltaQ-outlet
    OUT2: [kg/s] mp-outlet
    OUT3: [W] Pel
    OUT4: [kg/h] mp-outlet for plotting
    OUT5: [C] temp of pump
    OUT6: [Ws] sys gains
    OUT7: [Ws] sys losses
    OUT8: [Ws] sys capacity
    OUT12: [Pa]     a(ctr) adaptive coeff deltaP=a(ctr)+b(ctr)*Vp+c(ctr)*vp^2
    OUT13: [Pa/l/s] b(ctr)
    OUT14: []   c(ctr)
    OUT15: [Pa] deltaP is written by hydraulic unit

    DERIV1: [C] Temperature of pump
*/

/*----------------------------------------------------------------------------*/
int a=0; /* debug with a=1 */
int pedantic=0; /* !!! only use this flag if just one pump is existant !!! */
double Q_out=0,Tnew=0,Tin;
double mp_old=0;
double mp_max=par[unit][1];
double cp    =par[unit][2]*1e+3;
double rho   =par[unit][6]; /* used in mode 1 */
double Q_in  =out_o[unit][8]; /* get old energy from last timestep */
double mp_new=out_n IN(2);
double ctr   =out_n IN(3);

int first_time_flag=0; /* first call of unit in timestep */
/* int SWITCH_ON=0,SWITCH_OFF=0; / * check pump flipping */
int mode=(int) par[unit][5];
double mp_cp_h,mc,energy;
int pump_index=type_index[unit]; /* get id of pump */

if (select_unit[unit]++==0)  first_time_flag=1; 

if (rho==0) rho=1000; /* kg/m3 */
if (par[unit][7]<mp_max*h/3600) { /* min one plug capacity  mp_max*h/3600!!! */
  if (init_flag) printf("pump:%d dynamic mass default:%.1f [kg] \n",
                        unit,mp_max*h/3600);
  mc=mp_max*h*cp/3600; /* total capacity 10kg  */
  }
else   mc=par[unit][7]*cp; 

if (mp_new*3600>mp_max*1.1) {
  printf("pump:[%d] mp_new:%.1f > mp_max:%.1f=par1 \n",unit,mp_new*3600,mp_max);
  exit (-1);
}

/*-----------INIT-------------------------------------------------------------*/
if (init_flag) { /* default temp from input5 initial */
  /* thermo_struct definition, see 'global.orig.h' */
  thermo_struct[unit]=malloc_thermo_struct(pump_index); /* alloc just 1 field */
  if (cp<0) (*thermo_struct[unit]).fluid_property_class=(int) (-1*PAR (2));
  (*thermo_struct[unit]).pump_unit=unit;
  (*thermo_struct[unit]).cp=Get_cp_from_Temp(unit,METHOD_WATER,20 /*T*/ );
  (*thermo_struct[unit]).fluid_property_class=METHOD_WATER;

  out_o[unit][5]=out_n[unit][5]=out_n[unit][OUT_MAX0+5]; 
  if (out_o[unit][5]==0.0)  {
    printf("default ambient temperature for pump unit:%d Tinit%.1f\n",
           unit,20.0); 
    out_o[unit][5]=out_n[unit][5]=20;
  }
  deriv_o[unit][1]=deriv_n[unit][1]=out_o[unit][5];
  init_energy[unit][1]=deriv_o[unit][1]*mc;
}
if (0) printf("tsim:%.1f thermo_struct[%d].fluid_property_class:%d cp:%.2f\n"
         ,tsim,unit,(*thermo_struct[unit]).fluid_property_class
         ,(*thermo_struct[unit]).cp);

/*---mode 1 mpset by inp4 ----------------------------------------------------*/
if (  ( (mode==2) && ( in_u[unit][4]!= unit) )  /* mode2 and inp is connected */
    ||  (mode==1) /*  mode 1 */
   )
   {                                  /* set mp with inp4 */
   mp_new = out_n IN(4)/3600; /* set mp kg/s */
   /* if (rho!=0) mp_new*=rho*1e-3; / * kg/m3 --> kg/l */
   }
else {  /* use ctr-input 3  , mode 0 : mp=ctr*mp_max */
  mp_new = par[unit][1]/3600*out_n IN(3);
  if ( (ctr>1) || (ctr<0) ) { 
    printf("pump[%d] ctr:%.1f error\n",unit,ctr); 
    exit (-1); 
  } 
}

mp_cp_h=mp_new*cp*h; /* input-plug */

if (a) printf("\nmp_set PUMP:%d =%.1f kg/h\n",unit,out_n[unit][2]*3600);


/*--------------- first call of pump  ----------------------------------------*/
             /* maybe mp_out is not ready: previous_unit!=in_u[unit][1])  */
if (first_time_flag) /* bug 5/97 if previous unit of pump is heatexch     */
  {
  if (a) printf("PUMP:%d first loop: mp_old:%.2f mp_new:%.2f\n"
                ,unit,mp_old*3600,mp_new*3600);

  Q_out=mp_cp_h*deriv_o[unit][1]; /*Q_fluid_out */
  Q_in =0; /* Qfluid in */
  qp_sum[unit][1][1]+=Q_out;
 
  out_n[unit][1]=Q_out;
  out_n[unit][2]=mp_new;

  return (1);
  } /* end of first call */
/*------------------------- second loop --------------------------------------*/
else {
  if (a) printf("PUMP:%d second loop: mp_change mp_old:%.2f mp_new:%.2f\n"
                ,unit,mp_old*3600,mp_new*3600);

  Q_in=out_n IN(1);    /* get dQ in from closed loop */
  Q_out=out_n[unit][1]; /* calc at 1st call */
  /*qp_sum[unit][1][1]+=Q_out;*/
  qp_sum[unit][1][2]+=Q_in; 
  energy=init_energy[unit][1]
             -qp_sum[unit][1][1]   /* Q_fluid_out */
             +qp_sum[unit][1][2];  /* Q_fluid_in  */
  deriv_n[unit][1]=energy/mc;
  Tnew=deriv_n[unit][1];
  if ( (pedantic) && (fabs(Tnew-deriv_o[unit][1])>10) ) 
     printf("pump:%d (dynamic) Told:%.1f Tnew:%.1f\n",unit
                        ,deriv_o[unit][1],Tnew);
  if (Q_in!=0) Tin=Q_in/mp_cp_h;
  else         Tin=Tnew;

  out_n[unit][1]=Q_out;
  out_n[unit][2]=mp_new;

  out_n[unit][5]=Tnew;
  
  out_n[unit][6]=0; /* sys gains */
  out_n[unit][7]=0; /* sys loss */
  out_n[unit][8]=energy; /* sys cap */

  if (Q_in!=0) out_n[unit][12]=Q_in/(mp_cp_h);   /* TempIn */
  else         out_n[unit][12]=Tnew;
           
  if  ( (mp_new==0) && (Q_in!=0) ) 
    printf("pump.c:%d second loop mp_new==0 && Q_in:%.1f\n",unit,Q_in); 
                 
  if ( (Q_in==0) && (mp_new!=0) )
    { printf("pump:%d  second loop Q_in=0! mp_new:%.1f kg/h \n"
            ,unit,mp_new*3600);
      exit (-1);
    }
  if (a) printf("PUMP:%d second loop: Q_in:%.2f\n ",unit,Q_in);

/*-----------------------------check mp in&out-messflow balance---------------*/
  if (  ( (fabs(out_n[unit][2]-out_n IN(2) ))>1e-10 ) && (!init_flag)) {
     printf("pump:%d  mp-balance error diff:mp_in-mp_out:%e\n"
            ,unit,out_n[unit][2]*3600-out_n IN(2) *3600);
     exit (-1);
  }

} /* end of second loop */     

                 
/*-----------------------set output---second loop-----------------------------*/
  /* this type is called twice per simulation timestep */
  
  /* out 3 P_el = Pmax*mp/mp_max   Pel=(mp/mp_max)**a*Pmax */
out_n[unit][3]=pow((out_n[unit][2]*3600/par[unit][1])
                    ,par[unit][4])*par[unit][3];

out_n[unit][4]=out_n[unit][2]*3600; /* kg/h */
out_n[unit][5]=out_o[unit][5]=Tin;  /* rewrite current temp */

if ( (pedantic) &&(!first_time_flag) ){ /* unit energy check */
 static double Qunit_transfer_sum=0,Qinit=0,Qsys_losses_sum=0;
 double Qunit_transfer,Qsys_losses,energy_change;

 energy_change=out_n[unit][8]-out_o[unit][8];
 Qunit_transfer= out_n IN (1)-out_n[unit][1];
 Qsys_losses=0;
 if ( ( fabs(Qunit_transfer-Qsys_losses-energy_change)>1)&&(!init_flag) ) {
     printf("pump: energy_check error tsim:%.1f Qunit_transfer:%.1f "
             "Qsys_losses:%.1f energy_change:%.1f error:%.1f \n",
             tsim,Qunit_transfer,Qsys_losses,energy_change,
             Qunit_transfer-Qsys_losses-energy_change);
     printf("in1:%.1f out1:%.1f \n", out_n IN (1),out_n[unit][1]);
     exit (-1);
 } /* error */
    /* history energy check */
 if (init_flag) Qinit=out_n[unit][8]+Qsys_losses;
 Qunit_transfer_sum+=Qunit_transfer;
 Qsys_losses_sum+=Qsys_losses;
 if ( (fabs(Qinit+Qunit_transfer_sum-Qsys_losses_sum-out_n[unit][8])>1) 
      && (!init_flag) ) {
    printf("pump: history energy_check error tsim:%.1f Qinit:%.1f "
           "Qunit_transfer_sum:%.1f "
           "Qsys_losses_sum:%.1f sys_energy:%.1f error:%.1f\n",
             tsim,Qinit,Qunit_transfer_sum,Qsys_losses_sum,out_n[unit][8],
             Qinit+Qunit_transfer_sum-Qsys_losses_sum-out_n[unit][8]);
    exit (-1);
 } /* error */
} /* unit energy check */

return 1;
} /* pump() */
/*----------------------------------------------------------------------------*/





