/* read_data.c */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/

/*
  equidistant timestep data reader
  time profils with user defined timeintervalls (par1) are generated
  at outputs correspondating to the absolute time of year
*/
#include "global.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>


/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/


/*--------------read_data.c --TYPE 37 -------------------------------------*/

int read_data(tsim,unit,par,out_o,out_n,year_start,month,day,day_of_year)
              double tsim;    /* simulation time in seconds of this day */
             int     unit;    /* current unit number */
             double **par;    /* parameter array par[unit][1..PARMAX] */
             double **out_o;  /* output array with privious state of out[][OUTMAX] */
             double **out_n;  /* output array with current state */
             int year_start;  /* filename generate */
             int month;
             int day;
             int day_of_year;   /* int day_of_year */
{

/*  etc/xfig.awk

    TYPE 37
    PRIORITY  1
    HYDRAULIC 99


    PAR1: 3600 [s] time intervall of data base
    PAR2: 365  [d] days max of data base
    PAR3: 274  [d] starting day of database , default 0 

    OUT0: lines_max
    OUT1: [] column1
    OUT2: [] column2
    OUT3: [] column3
    OUT4: [] column4
 */
                                                            
#define CHAR_MAX 300
#define COLUMNS_MAX 30

int a=0; /*-----------debug-----*/
int words_max=0,i,i2,i_line=0;
int table_columns=0;
FILE *fd;
char filename[40]={"./projects/NEGEV/data/database.txt"};
char ascii_line[CHAR_MAX]={"\0"};
static int init_flag=1;
static double **data_field;
int word_pointer=0;
char word[COLUMNS_MAX][CHAR_MAX];

double time_intervall=par[unit][1];
double days_max      =par[unit][2];
double starting_day  =par[unit][3];
int current_line;
int data_line   =(int) out_o[unit][0];
int columns_max =(int) out_o[unit][OUT_MAX0-1];



if (init_flag)
 {
 init_flag=0;
 data_line=0;
 printf("read_data: reading data:%s\n",filename);
 if ((fd=fopen(filename,"r"))==NULL) 
   { printf("read_data:fopen:%s error\n",filename); exit (-1); };

 /* scan lines to check columns */
 i_line=0;
 while (fgets(ascii_line, CHAR_MAX ,fd)!=NULL)
  {
  words_max=0;
  for (i=1;i<strlen(ascii_line);i++) 
    if ( (ascii_line[i]==' ') && (ascii_line[i-1]!=' ') ) words_max++; 
  if (i_line==0) /* check headline */
    {
    table_columns=words_max;
    if ( (ascii_line[0]=='#') && (ascii_line[1]==' ') ) table_columns--; /* '# TIME' */
    if (ascii_line[0]!='#') {printf("read_data: no headline with '#'\n"); exit(-1); }
    }
  if (0) printf("words_max:%d table_columns:%d\n",words_max,table_columns);     
  if ( (words_max!=table_columns) && (ascii_line[0]!='#') ) {
    printf("read_data:headline_columns different from words of this line:%d\n"
           ,i_line); 
    printf("line:%s\n",ascii_line);
          exit (-1); 
     
    }
  i_line++;
  if (ascii_line[0]!='#') data_line++;
  } /* while() read lines */

 if (a) printf("file:%s lines:%d columns:%d \n",filename,data_line,words_max); 
 data_field=malloc_2double(table_columns+1,data_line+1);
 if (a) printf("read_data: malloc done \n");

 /* read words to data array */
 i_line=0;
 rewind(fd); /* back to begin */
 data_line=0;
 while (fgets(ascii_line, CHAR_MAX ,fd)!=NULL)
  {
  if (ascii_line[0]!='#') /* select data lines */
   {
   words_max=0;word_pointer=0;i2=0;
   for (i=1;i<strlen(ascii_line);i++) 
     {
     if ( (ascii_line[i]==' ') && (ascii_line[i-1]!=' ') ) 
       {
       if (0) printf("i:%d,",i);
       words_max++; word_pointer=0;i2=0;
       }
     else 
       { 
       word[words_max][word_pointer]=ascii_line[i];
       word[words_max][word_pointer+1]='\0';
       word_pointer++; 
       }
     }
     for (i=0;i<=words_max;i++)
       {
       data_field[i][data_line]=strtod(word[i],NULL);
       if (a) printf("%d:%.4f ",i,data_field[i][data_line]);
       }
     if (a) printf("\n data_line:%d: ",data_line);
     
   data_line++;
   }
  } /* while() read lines */
 if (a) printf("\n");
 out_n[unit][0]=out_o[unit][0]=data_line;
 out_n[unit][OUT_MAX0-1]=out_o[unit][OUT_MAX0-1]=columns_max=(double) words_max;
  
 fclose(fd);
 
 if (data_line*time_intervall<days_max*3600*24) /* data_line 8760 */
  {
  printf("read_data[%d]:data_line*time_intervall=%.1f<days_max*24*3600:%.1f\n"
         ,unit,data_line*time_intervall,days_max*24*3600); exit (-1);
  }
 } /* end init_flag */


current_line=(int) ( (double) ((day_of_year-1)*24*3600+tsim)/ (double) time_intervall);
if (starting_day>0) 
 {
 if (day_of_year>=starting_day) 
  current_line=(int) ( (double) ((day_of_year-starting_day)*24*3600+tsim)
                     / (double) time_intervall);
 else 
  current_line=(int) ( (double) ((day_of_year+days_max-starting_day)*24*3600+tsim)
                     / (double) time_intervall);
 }
if (a) printf("tsim:%.1f day_of_year:%d: hour:%d sec_of_year:%d current_line:%d\n"
   ,tsim,day_of_year,day_of_year*24,day_of_year*24*3600,current_line);


if (current_line>data_line) 
 { printf("current_line:%d\n",current_line); exit(-1); }

for (i=1;i<=columns_max+1;i++) out_n[unit][i]=data_field[i-1][current_line];
if (a) printf("tsim:%.1f current_line:%d day_of_year:%d out1:%.1f out:%d:%.1f\n"
   ,tsim,current_line,day_of_year,out_n[unit][1],columns_max+1
   ,out_n[unit][columns_max+1]);


return (1);
}


