/* storage2.c */
/*       10        20        30        40        50        60        70       */
/*3456789012345678901234567890123456789012345678901234567890123456789012345671*/
/*----------------------------------------------------------------------------*/

#include "global.h" 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
000411 chriwi update function call, backup in src/history
       no i_dead_nr for pipe calls any more !! not checked
       malloc of internal pipe_units , new index
       
*/
/*-----------Storage2 ----------------------------------------*/
int storage (unit,par,in_u,in_nr,out_n,out_o,deriv_o,deriv_n,h,
             init_energy,qp_sum,tsim,init_flag,type,day_of_year,
             previous_unit,un_max)
           int    unit;      /* current unit number */
           double **par;     /* parameter array par[unit][1..PARMAX]  */
           int    **in_u;    /* input unit array, input connections   */
           int    **in_nr;   /* input number array                    */
           double **out_o;   /* output array with privious state      */
           double **out_n;   /* output array with current state       */
           double **deriv_o; /* derivatives: used for diffential equ. */
           double **deriv_n; /*  new state of deriv                   */
           double h;         /* sim time step                         */
           double **init_energy; /* initial energy in RC-comp. [Ws]   */
           double ***qp_sum; /* energy to one node qp[unit][n][m]     */
           double tsim;      /* simulation time in [s] of this day    */
           int    init_flag; /* set at initial loop of sim            */
           int    *type;     /* type number array                     */
           int    day_of_year;
           int    previous_unit;
           int    un_max;    /* unit max */

{
/* xfig.awk             

   TYPE 04
   PRIORITY  6          used by config routine
   HYDRAULIC 4          4-pole fluid            


   PAR1: 10 [W/K] kA ambient losses
   PAR2: 1000 [l] volume incl heatexchangers
   PAR3: 4.19 [kJ/kgK] cp of storage fluid 
   PAR4: 10 [] number n of nodes
   PAR5: 3 [] inlet  node of WT1 !! set par6=par5 !!,set by inp11
   PAR6: 3 [] outlet node of WT1,set by inp11
   PAR7: 4.19 [kJ/kgK] cp in WT1
   PAR8: 4 [m] l1 length
   PAR9: 15 [mm] d1 diameter [mm]
   PAR10: 50 [W/Km] kA1/l linear heat transfer coeff, no fit
   PAR11: 6 [] inlet  node of WT2 !! set par11=par12 !!
   PAR12: 9 [] outlet node of WT2 
   PAR13: 4.19 [W/kgK] cp in WT2
   PAR14: 4 [m] l2
   PAR15: 15 [mm] d2
   PAR16: 50 [W/Km] kA2/l, 0: internal FIT with T_wt2_amb wt2_out
   PAR17: 1 [] node sensor 1
   PAR18: 2 [] node sensor 2
   PAR19: 8 [] node sensor 3
   PAR20: 10 [] node sensor 4
   PAR21: 3 [] circulation node inlet
   PAR22: 0 [] mode 0:2WT 1:0WT 2:1WT (solar)
   PAR23: 1 [] circulation node outlet,if 0-->par23=1
   PAR24: 10 [] WW node back (CW in ) ,if 0 then par24=par4,set by inp11
   
   INP1: 0 [Ws] WT2: deltaQ_kollinlet
   INP2: 0 [kg/s] WT2: mp_kollinlet
   INP3: 0 [Ws] deltaQ_cw
   INP4: 0 [kg/s] mp_cw [kg/s]
   INP5: 0 [Ws] WT1: dQ_heater
   INP6: 0 [kg/s] WT1: mp_heater
   INP7: 0 [Ws] circulation_in: dQ_ci
   INP8: 0 [kg/s] circulation_in: mp_ci
   INP9: 20.0 [C] ambient temperature
   INP10: 50.0 [C] internal loop, from out13 
   INP11: 0 [] mode of operation: 0:HW_load,1:Circ_ext,2:warm-upHW,3:heating

   OUT1: [Ws] he2: deltaQ_kolloutlet
   OUT2: [kg/s] he2: mp_kolloutlet
   OUT3: [Ws] deltaQ_ww
   OUT4: [kg/s] mp_ww
   OUT5: [] i_dead_nr
   
   OUT6: [Ws] system inlet energy 
   OUT7: [Ws] system outlet energy 
   OUT8: [Ws] internal energy
   
   OUT9: [Ws]  he1: deltaQ_heater OUT
   OUT10: [kg/s] he1: mp_heater
   OUT11: [C] T_sensor1 ww node PAR17
   OUT12: [C] T_sensor2    node PAR18
   OUT13: [C] T_sensor3    node PAR19
   OUT14: [C] T_sensor4    node PAR20
   OUT15: [C] Tmean nodes he1
   OUT16: [C] Tmean nodes he2
   OUT17: [Ws] Q_ci_outlet
   OUT18: [kg/s] mp_ci_outlet
   OUT19: [] first_execute_time-->first time call of unit   
   OUT20: [W] Power heatexch. he2
   OUT21: [W] Power heatexch. he1
   OUT22: [W] Power of load_ww
   OUT23: [W] Power of losses to ambient
   OUT24: [C] temp wt2 out 
   OUT25: [Ws] internal energy of wt2
   OUT26: [C] temp wt2 (mean)
   OUT27: [W] Power of circ
   
   OUT31..39: [C] Temp node 1 ... 9 
   OUT[OUTMAX0+1]: [C] T_cw_in, if input1=[0,0] input initial
   OUT[OUTMAX0+2]: [kg/s] mp_cw  , if input1=[0,0] input initial
   ..

   DERIV1: 60 [C] temperatur of node 1 (top) init_energy 1=DERIV1/mc
   DERIV2: 60 [] node 2
   DERIV3: 60 []
   DERIV4: 60 []
   DERIV5: 60 [] node 5 lowest
   DERIV(un_max+nr  ): 60 [C] he1 temp
   DERIV(un_max+nr-1): 60 [C] he2 temp

*/

  int a=0; /*-------------ANALYSE a=1 ---------*/
  int pedantic=0;
  static int storage_counter=0; /* define number of storages */
  static int storage_selector[STORAGE_MAX]; /* current unit number */
  static double **pipe_par;
  static int    **pipe_in_u;
  static int    **pipe_in_nr;
  static double **pipe_deriv_o;
  static double **pipe_deriv_n;
  static double **pipe_out_o;
  static double **pipe_out_n;
  static double **pipe_init_energy;
  static double ***pipe_qp_sum;

  int conv_flag; /* stratisfaction & mixing */
  int init_conduction=1;
  int mode=(int) par[unit][22]; /* 0 [] mode 0:2WT 1:0WT 2:1WT (solar) */
  double cp=par[unit][3]*1e+3;
  double mc_node; /* vol*density*cp of storage water  */
  double mp_cp_KW  =out_n[  in_u[unit][4] ]
                         [ in_nr[unit][4] ]*cp;
  double mp_cp_circ=out_n[  in_u[unit][8] ]
                         [ in_nr[unit][8] ]*cp;
  double mp_cp_WT1 =out_n[  in_u[unit][6] ]
                         [ in_nr[unit][6] ] *par[unit][7]*1e+3;
  double mp_cp_WT2 =out_n[  in_u[unit][2] ]
                         [ in_nr[unit][2] ] *par[unit][13]*1e+3;
  double Q_help,Q_sum=0;
  double Q_out,Q_out2; /* energy input from wt1 & wt2 */
  int nr;
  double first_execute_time=out_n[unit][19];

  int n=(int) par[unit][4];       /* number of nodes */
  int wt1_in =(int) par[unit][5];  /* wt1_in node of inlet wt1 */
  int wt1_out=(int) par[unit][6];  /* oulet-node of WT1 */
  double d1=par[unit][9]; /*diameter wt1 */
  double l1=par[unit][8]; /*length of WT1 */

  int wt2_in =(int) par[unit][11];  /* inlet-node of wt2 */
  int wt2_out=(int) par[unit][12]; /* oulet-node of WT2 */
  double d2=par[unit][15]; /*diameter wt2 */
  double l2=par[unit][14]; /*length of WT2 */
  double Tin_wt2,T_amb_wt2;  
  double Vol1,Vol2,mp_wt2,kA_wt1,kA_wt2;
  int step_wt1,step_wt2,i,i1;
  int call_wt1_flag,call_wt2_flag,circ_flag,KW_flag;
  int first_time_flag; /* first call of unit in timestep */
  double kA_node=par[unit][1]/n;
  double energy[DERIV_MAX];
  double Qi[NODE_MAX][10];


  if   (first_execute_time!=tsim) {
                                  out_n[unit][19]=tsim;
                                  first_time_flag=1;
                                  }
  else                            first_time_flag=0;
  
  if (a) printf("--------------------------------------------------------"
                "----UNIT:%d tsim:%.1f-\n",unit,tsim);

  mp_wt2=out_n[  in_u[unit][2] ]
              [ in_nr[unit][2] ];
              
  /* !!! that's no fine way to set parameters, but sometimes ... !!!!!!!!! */
  switch ( (int) out_n[  in_u[unit][11] ]
                      [ in_nr[unit][11] ] )  /* operation mode */
         {
         case 0: par[unit][24]=10; /* WW back node, CW inlet */
                 break;
         case 1: par[unit][24]= 5; /* WW back node CW inlet for circ */
                 break;
         case 2: par[unit][5 ]=1;  /*  WT1 inlet  2:warm-upHW */
                 par[unit][6 ]=3;  /*  WT1 outlet */
                 break;
         case 3: par[unit][5 ]=4;  /*  WT1 inlet  for HW 3:heating */
                 par[unit][6 ]=7;  /*  WT1 outlet */
                 break;
         case 4: break; /* no change */
         case 5: /* WT2 is stratified */
                 if (mp_wt2>0) Tin_wt2=out_n[  in_u[unit][1] ]
                                            [ in_nr[unit][1] ]
                                       /( mp_wt2*par[unit][13]*1e+3*h);
                 else          Tin_wt2=deriv_n[unit][wt2_in];
                 i=par[unit][4]; /*last node */
                 do i--; while (   (i>1) && (Tin_wt2>deriv_o[unit][i]) );
                 par[unit][11]=i; /* inlet node of WT2 direct */
                 par[unit][12]=par[unit][4]; /* last node */
                 break; 
         default: printf("error in operation mode unit:%d\n",unit);exit (-1);
         }
     /* LowFlow Valve */
   if (in_u[unit][11]!=unit) /* ctr inp11 set */
     {
     if (out_n[  in_u[unit][11] ]
              [ in_nr[unit][11] ]-(int)  out_n[  in_u[unit][11] ]
                                              [ in_nr[unit][11] ] >= 0.05 ) 
            par[unit][11]=1; /* WT2 inlet at top node */
     else     { par[unit][11]=6; 
               /*if (out_n[  in_u[unit][2] ][ in_nr[unit][2] ] >0) 
                  printf("now node 6 inlet T6:%.1f \n",deriv_o[unit][6]); */}
     }
            

  /* select input-unit for storage (storage is a multi-called unit !!) */
  call_wt1_flag=(    (wt1_in)
                  && (in_u[unit][5]==previous_unit)
                )
                ||
                (    (in_u[unit][5]==unit)
                  && (first_time_flag)
                );

  call_wt2_flag=(    (wt2_in)
                  && (in_u[unit][1]==previous_unit)
                )
                || 
                (    (in_u[unit][1]==unit)
                  && (first_time_flag)
                );
               
  circ_flag=    (     in_u[unit][7]==previous_unit)
                || 
                (    (in_u[unit][7]==unit)
                  && (first_time_flag)
                );

  KW_flag  =    (     in_u[unit][3]==previous_unit)
                || 
                (    (in_u[unit][3]==unit)
                  && (first_time_flag)
                );

  /* Volume of heat-exchangers WTs in [l] */
  Vol1=3.1415*d1*d1/4*l1*1e-3;
  Vol2=3.1415*d2*d2/4*l2*1e-3;

  /* heat-capacity */
  mc_node=(par[unit][2]-Vol1-Vol2)*cp/n; /* rho:=1 */

  if (a) printf("mc_node: %.4f \n",mc_node);

  if (wt1_out!=wt1_in) step_wt1=(wt1_out-wt1_in)/(abs(wt1_out-wt1_in));
  else                 step_wt1=1;
  if (wt2_out!=wt2_in) step_wt2=(wt2_out-wt2_in)/(abs(wt2_out-wt2_in));
  else                 step_wt2=1;

      /* if (wt1_out==wt1_in)
       kA_wt1=par[unit][10]*par[unit][8]; / * kA-total * /
      else
       kA_wt1=par[unit][10]*par[unit][8] /(abs(wt1_out-wt1_in));
      */
  kA_wt1=par[unit][10]*par[unit][8]; /* kA-total */
  kA_wt2=par[unit][16]*par[unit][14]; /* kA-total */


  /*---------- Initialize ------------------------------------------------- */
  if ( (init_flag) && ( !(int) out_o[unit][5]) )
    {
      int i;
      static int init_flag_local=1;
      double q_storage=0,T;

    if (init_flag_local) { /* install virtual pipe array */
          /* malloc anything */
        pipe_par    =malloc_2double(STORAGE_MAX*2+1,PAR_MAX+1);
        pipe_in_u   =malloc_2int   (STORAGE_MAX*2+1,INPUT_MAX+1);
        pipe_in_nr  =malloc_2int   (STORAGE_MAX*2+1,INPUT_MAX+1);
        pipe_deriv_o=malloc_2double(STORAGE_MAX*2+1,DERIV_MAX+1);
        pipe_deriv_n=malloc_2double(STORAGE_MAX*2+1,DERIV_MAX+1);
        pipe_out_n  =malloc_2double(STORAGE_MAX*2+1,OUT_MAX+1);
        pipe_out_o  =malloc_2double(STORAGE_MAX*2+1,OUT_MAX+1);
        pipe_init_energy=malloc_2double(STORAGE_MAX*2+1,DERIV_MAX+1);
        pipe_qp_sum=malloc_3double(STORAGE_MAX*2+1,DERIV_MAX+1,PATH_MAX+1);

        init_flag_local=0; /* only once */
        }
      if (first_time_flag) { storage_counter++; 
                             storage_selector[storage_counter]=unit;
                             storage_selector[storage_counter+1]=0;
                             if (storage_counter>1) 
                                {
                                printf("storage: storage_counter:%d\n"
                                         ,storage_counter);
                                exit (-1);
                                }
                           }

      first_time_flag=1; /* tsim=out19=0 */
      if ( (int) par[unit][24]==0) {
          par[unit][24]=par[unit][4]; /* set CW node in to n */
          printf("par[%d][24] set to %d\n",unit,(int) par[unit][4]);
          }
      if ( (int) par[unit][23]==0) { par[unit][23]=1; /* set CW node in to n */
           printf("par[%d][23] set to %d\n",unit,(int) par[unit][23]);   }

      

      /*------ check WT-parameter */
      if (   (wt1_in >n) || (wt1_in<1)
           ||(wt1_out>n) || (wt1_out<1)
           ||(wt2_in >n) || (wt2_in <1)
           ||(wt2_out>n) || (wt2_out<1)
         )
         {
           printf("exit, caused by WT-Parameters\n");
           exit(-1);
         }

      if ( deriv_o[unit][1]==0 )
         printf(" set deriv1 of storage unit:%d",unit);

      /*------ init energy */
      T=10; /* default temperature */
      for (i=1;i<=n;i++)
        {
          if ( deriv_n[unit][i]>0) T=deriv_o[unit][i]=deriv_n[unit][i];
          else deriv_o[unit][i]=deriv_n[unit][i]=T; /* at least deriv1!=0 */
          init_energy[unit][i]=T*mc_node;
          q_storage+=init_energy[unit][i];

          if (a) printf("deriv_o: %.4f  deriv_n: %.4f \n"
                        ,deriv_o[unit][i]
                        ,deriv_n[unit][i]
                        );
        };

      if (wt1_in) 
        {
         pipe_par  [storage_counter*2+0][1]=l1;   /* length of pipe wt1 */
         pipe_par  [storage_counter*2+0][2]=d1;   /* diameter wt1 */
         pipe_par  [storage_counter*2+0][3]=par[unit][7];
         pipe_par  [storage_counter*2+0][4]=kA_wt1/l1;  /* kA/l wt1 */
         
         pipe_in_u [storage_counter*2+0][1]=in_u [unit][5]; /* dQ_wt1_in */
         pipe_in_nr[storage_counter*2+0][1]=in_nr[unit][5];
         pipe_in_u [storage_counter*2+0][2]=in_u [unit][6]; /* mp_in */
         pipe_in_nr[storage_counter*2+0][2]=in_nr[unit][6];
         pipe_in_u [storage_counter*2+0][3]=unit;      /* connect node temp */
         pipe_in_nr[storage_counter*2+0][3]=15; /* T node wt1_in: T_wt1_amb */

         /* outs set by pipe(); */

         /* set T node wt1_in for pipe-proc */
         out_n[unit][15]=out_o[unit][15]=deriv_n[unit][wt1_in];
         out_o      [ pipe_in_u [storage_counter*2+0][3] ]
                    [ pipe_in_nr[storage_counter*2+0][3] ]
              =out_n[ pipe_in_u [storage_counter*2+0][3] ]
                    [ pipe_in_nr[storage_counter*2+0][3] ]
              =out_n[unit][15];

         pipe_deriv_o     [storage_counter*2+0][1]
             =pipe_deriv_n[storage_counter*2+0][1]=deriv_o[unit][wt1_in];
        }

      /* now wt2 */
      if (wt2_in)
        {
         pipe_par  [storage_counter*2+1][1]=l2;   /* length of pipe wt2 */
         pipe_par  [storage_counter*2+1][2]=d2;   /* diameter wt2 */
         pipe_par  [storage_counter*2+1][3]=par[unit][13];
         pipe_par  [storage_counter*2+1][4]=kA_wt2/l2;  /* kA/l wt2 */
         

         pipe_in_u [storage_counter*2+1][1]=in_u [unit][1]; 
         pipe_in_nr[storage_counter*2+1][1]=in_nr[unit][1];
         pipe_in_u [storage_counter*2+1][2]=in_u [unit][2]; 
         pipe_in_nr[storage_counter*2+1][2]=in_nr[unit][2];
         pipe_in_u [storage_counter*2+1][3]=unit;      /* connect node temp */
         pipe_in_nr[storage_counter*2+1][3]=16; /* T node wt1_in: T_wt1_amb */

         /* outs set by pipe(); */

         /* set T node wt1_in for pipe-proc */
         out_n[unit][16]=out_o[unit][16]=deriv_n[unit][wt2_in];
         out_o      [ pipe_in_u [storage_counter*2+1][3] ]
                    [ pipe_in_nr[storage_counter*2+1][3] ]
              =out_n[ pipe_in_u [storage_counter*2+1][3] ]
                    [ pipe_in_nr[storage_counter*2+1][3] ]
              =out_n[unit][16];


         pipe_deriv_o     [storage_counter*2+1][1]
             =pipe_deriv_n[storage_counter*2+1][1]=deriv_o[unit][wt2_in];


        }


      /*------- set sensor outs */
      out_o[unit][11]=deriv_o[unit][ (int) par[unit][17] ]; /* T_WW */
      out_o[unit][12]=deriv_o[unit][ (int) par[unit][18] ];
      out_o[unit][13]=deriv_o[unit][ (int) par[unit][19] ];
      out_o[unit][14]=deriv_o[unit][ (int) par[unit][20] ]; /* T_KW */

      if (a) printf("T_sensor1 WW:%.4f\n",out_o[unit][11]);
      if (a)
        printf("init-energy of heat storage: %.2f \n",
                q_storage/3.6e+6);

    };
  /* ----------------------------*/



  /* energy input caused by heatexchangers */
  if ( (call_wt1_flag)  && (par[unit][22]==0) ) /* mode 0 2 WT */
    {

    pipe2(storage_counter*2+0,pipe_par,pipe_in_u,pipe_in_nr,pipe_out_o,
          pipe_out_n,pipe_deriv_o,pipe_deriv_n,h,pipe_init_energy,pipe_qp_sum,
          tsim,init_flag);
    /*
    pipe2(un_max+nr,par,in_u,in_nr, out_o,out_n, deriv_o,deriv_n,
        h,init_energy,qp_sum, tsim,init_flag);
    */

    Q_out=pipe_out_n[storage_counter*2+0][12]; /* dQ_wt1-->node wt1_in..wt1_out */

    }
  else Q_out=0;

  if ( (call_wt2_flag) && (par[unit][22]!=1) ) /* mode1: 0 WT-->WT2 existent */
    {

      
       /* logarithmic FIT
          KAL = ln(KA)
          KALFIT   =A[1]+A[2]*ln(MP1)+A[3]*ln(TIN1-TPU)+A[4]*ln(0.5*(TIN1+TPU))
          KAF=exp(A[1])*MP1^A[2]*(TIN1-TPU)^A[3]*(0.5*(TIN1+TPU))^A[4]
          -8.71288989e-01,8.79814408e-01,6.02833996e-02,3.24111543e-01
          linear FIT:
          KAFIT = A[1] + A[2]* TIN1 + A[3]* TPU +A[4]*MP1
          -2.74765394e+01,5.04730386e-01,2.89605527e-01,8.26217811e-01
          FIT: par  [un_max+nr-1][4]=kA_wt2/l2; 
          mp=250kg/h linear fit 3D-fit-linear.ft 
          par  [un_max+nr-1][4]=(105.081920+2.0586*Tin_wt2-0.1079*T_amb_wt2)/l2;
          mp=50kg/h     
          par  [un_max+nr-1][4]=(4.63226245e+01+1.15765287e-1*Tin_wt2
                                +6.89941290e-02*T_amb_wt2)/l2;
          FIT lin 50 kg/h par  [un_max+nr-1][4]=
                   (4.64934767e+01+1.34104186e-01*Tin_wt2
                   +4.07837655e-02*T_amb_wt2)/l2;
          FIT log mp 50 ..250    
          par  [un_max+nr-1][4]
             = fabs( (exp(-8.71288989e-01)*pow(mp_wt2*3600,8.79814408e-01)
                     *pow( (Tin_wt2-T_amb_wt2),6.02833996e-02)
                     *pow( (0.5*(Tin_wt2+T_amb_wt2) ),3.24111543e-01))/l2 
                  );   
          lin for all points 107.29+2.2599*Tin_wt2-0.5229*T_amb_wt2;
        */        
        
     if (mp_wt2>0) Tin_wt2=out_n[  in_u[unit][1] ]
                                [ in_nr[unit][1] ]/( mp_wt2*par[unit][13]*1e+3*h);
     else Tin_wt2=deriv_n[unit][wt2_in];

     T_amb_wt2= deriv_n[unit][wt2_in]; /* out_o[unit][16];  Tmean wt2_in..wt2_out */
         
     if ( (int) par[unit][16] == 0) /* take FIT */
      {                                                         /* linear Fit */
        /* !!!!!!!!!!!!!!! factor 4.8 with UNISOL1500,also 4&6 lines later !! */        
        kA_wt2=4.8*(-2.74765394e+01 
               +5.04730386e-01*Tin_wt2
               +2.89605527e-01*T_amb_wt2
               +8.26217811e-01*mp_wt2*3600/4.8);
        if (kA_wt2<50) kA_wt2=50; /*limit U */

        pipe_par  [storage_counter*2+1][4]=kA_wt2/l2;
        if ( (mp_wt2*3600)>350*4.8) printf("storage u:%d mp_wt2>350kg/h -->FIT\n",unit);
        
      } 
     else /* take fixed heat coefficent */ 
      pipe_par  [storage_counter*2+1][4]=kA_wt2/l2; 
                                                          
     if (a)
      {
      if (mp_wt2*3600>200)
      printf("kA_wt2:%.1f Tin:%.1f  T_SP:%.1f Q_in:%.1f mp_wt2:%.1f kg/h\n"
             ,pipe_par  [storage_counter*2+1][4]
             ,Tin_wt2,T_amb_wt2,out_n[  in_u[unit][1] ]
                                     [ in_nr[unit][1] ],mp_wt2*3600);
      }
    /*
    pipe2(un_max+nr-1,par,in_u,in_nr, out_o,out_n, deriv_o,deriv_n,
        h,init_energy,qp_sum, tsim,init_flag);
    */

    pipe2(storage_counter*2+1,pipe_par,pipe_in_u,pipe_in_nr,pipe_out_o,
          pipe_out_n,pipe_deriv_o,pipe_deriv_n,h,pipe_init_energy,pipe_qp_sum,
          tsim,init_flag);


           /* shift outs */
           Q_out2=pipe_out_n[storage_counter*2+1][12];/*dQ_wt2-->node wt2_in..wt2_ou

           Q_out2=out_n[un_max+nr-1][12];/*dQ_wt2-->node wt2_in..wt2_out */
           if ( (T_amb_wt2<8) && (mp_wt2>0) && (Q_out2<-50) )
               printf("storage unit%d Q_out2:%.1f T_amb_wt2:%.1f kA_wt2:%.1f "
                      "Tin_wt2:%.1f\n"
                      ,unit,Q_out2,T_amb_wt2,kA_wt2,Tin_wt2);
           
    }
    else Q_out2=0;

 /* ---------energies for each layer----------------------------- */

  conv_flag=0; /*convection-flag */
  for (i=1;i<=n;i++)
    {
      /* now static heat conduction 2.node to wt1_out */
      if (call_wt1_flag)
      {
        if (   ( (i>=wt1_in) && (i<=wt1_out) && (step_wt1>0) )
            || ( (i<=wt1_in) && (i>=wt1_out) && (step_wt1<0) )
           )
         {
         if ( par[unit][22]==0 ) /* mode 0 2 WT */
          {
            if (i == wt1_out)      /*  last piece gets rest of energy */ 
              Qi[i][1]=Q_out-Q_sum;
            else
              {
              Qi[i][1]=Q_out/(abs(wt1_out-wt1_in)+1); /* Mean energy */
              Q_sum+=Qi[i][1];
              }
          }
         else /* mode !=0 */
          {
          if (i==wt1_in) 
               Qi[i][1]= out_n[  in_u[unit][5] ]
                              [ in_nr[unit][5] ] /* Qin WT1 */
                        -mp_cp_WT1*h*deriv_o[unit][i];
          else { if ( (wt1_in<=wt1_out) && (i<=wt1_out) )
                    Qi[i][1]= mp_cp_WT1*h*deriv_o[unit][i-1] /*top-down*/
                             -mp_cp_WT1*h*deriv_o[unit][i  ];
                 if ( (wt1_in>=wt1_out) && (i<=wt1_in) )
                    Qi[i][1]= mp_cp_WT1*h*deriv_o[unit][i+1] /*down-up*/
                             -mp_cp_WT1*h*deriv_o[unit][i  ];
               }
          }
         if (a) printf("wt1: Qi[%d][1]:%.4f \n",i,Qi[i][1]);
         }
         else Qi[i][1]=0;
      }
      else Qi[i][1]=0;

      /*-------------------------------- wt2----------------- */
      if (call_wt2_flag)
      {
        if (    ( (i>=wt2_in) && (i<=wt2_out) && (step_wt2>0) )
              ||( (i<=wt2_in) && (i>=wt2_out) && (step_wt2<0) )
           )
          {
          if ( (int) par[unit][22]!=1 ) /*  mode1: 0 WT-->WT2 existent */
            {
            if (i == wt2_out) Qi[i][8]=Q_out2;
            else              Qi[i][8]=0;
            
               /* if (i == wt2_out)      / *  last piece gets rest of energy * /
                 Qi[i][8]=Q_out2-Q_sum2;
                  else
                  {
                  Qi[i][8]=Q_out2/(abs(wt2_out-wt2_in)+1);
                  Q_sum2+=Qi[i][8];
                  }
               */  
            }
          else /* mode =1 */
           {
           if (i==wt2_in)
                Qi[i][8]= out_n[  in_u[unit][1] ] /* Qin WT2  */
                               [ in_nr[unit][1] ] -mp_cp_WT2*h*deriv_o[unit][i]; 
           else { if ( (wt2_in<=wt2_out) && (i<=wt2_out) )
                     Qi[i][8]= mp_cp_WT2*h*deriv_o[unit][i-1] /*top-down*/
                              -mp_cp_WT2*h*deriv_o[unit][i  ];
                  if ( (wt2_in>=wt2_out) && (i<=wt2_in) )
                     Qi[i][8]= mp_cp_WT2*h*deriv_o[unit][i+1] /*down-up*/
                              -mp_cp_WT2*h*deriv_o[unit][i  ];
                }
           }
           if (a) printf("wt2: Qi[%d][8]:%.4f \n",i,Qi[i][8]);
          }
          else Qi[i][8]=0;
      }
      else Qi[i][8]=0;

      /* -----------------------------------------------------loads */
      /* first mp_KW cold water inlet */
      if (KW_flag)
      {
         Qi[i][4]=Qi[i][5]=0; /* reset */
         if (i==( (int) par[unit][24] ))  /* not n any more */
          {
          Qi[i][4] = out_n[  in_u[unit][3] ]
                          [ in_nr[unit][3] ];   /* Q_KW_in */
          Qi[i][5]   =-mp_cp_KW*h*deriv_o[unit][i];  /* Q_WW */
          }
          if (  (i<(int) par[unit][24]) )    /* node  in */
          {
          Qi[i][4]   = mp_cp_KW*h*deriv_o[unit][i+1]; /* Q_KW_in */
          Qi[i][5]   =-mp_cp_KW*h*deriv_o[unit][i];  /* Q_WW */
          }
      }
      else Qi[i][4]=Qi[i][5]=0;

      /*------------------------------------------------- circulation */
      Qi[i][6]=Qi[i][7]=0; /* reset */
      if (circ_flag)
      {
       if (i==( (int) par[unit][21] ) ) /* Q_circ_inlet */
        {
          Qi[i][6]=out_n[  in_u[unit][7] ]
                        [ in_nr[unit][7] ]; /*dQ_circ */
          Qi[i][7]=-mp_cp_circ*h*deriv_o[unit][i];
        }
       else  /* bug 5/97 if par21<par23  */
        {                         
        if (   (    (int) par[unit][23] <        (int) par[unit][21])
            && ( i>=(int) par[unit][23]) && ( i< (int) par[unit][21])
           ) {
           Qi[i][6]= mp_cp_circ*h*deriv_o[unit][i+1]; /* Q_circ_in  */
           Qi[i][7]=-mp_cp_circ*h*deriv_o[unit][i  ]; /* Q_circ_out */
           }
        if (    (    (int) par[unit][23] >        (int) par[unit][21])
             && ( i<=(int) par[unit][23]) && ( i> (int) par[unit][21])
           ) {
           Qi[i][6]= mp_cp_circ*h*deriv_o[unit][i-1]; /* Q_circ_in  */
           Qi[i][7]=-mp_cp_circ*h*deriv_o[unit][i  ]; /* Q_circ_out */
           }
         }
         /* printf("circ_flag: u:%d mp:%.1f Qi[%d][6]:%.1f Qi[%d][7]:%.1f \n"
               ,unit,mp_cp_circ*3600/(4190),i,Qi[i][6],i,Qi[i][7]);
               */
         
      }

      if (first_time_flag)  /* calculate only once */
        {
        /*---------------------------- stratisfaction convection */
        if  ( (!conv_flag)
           && (deriv_o[unit][i+1]>deriv_o[unit][i])
           && (i < n)
          )
          {
            conv_flag=1; /* now evaluate energy flow in every following
                            layer */
            Qi[i][2]=0;
          }
        if (conv_flag)
          {
          if (    (deriv_o[unit][i+1]>deriv_o[unit][i])
               && ( i!=n )
             )
           {
             /* kA=40 W/m^2K FIT: 2000 W/K 500l tank */
             /* total mixing:  
              Q_help=mc_node*(deriv_o[unit][i+1]-deriv_o[unit][i]);
             */
             Q_help=4000*(par[unit][2]/500)*h*(deriv_o[unit][i+1]-deriv_o[unit][i]);
             
             Qi[i]  [2] += Q_help;
             Qi[i+1][2] = -Q_help;
           }
          else
           {
             Q_help=0;
             Qi[i+1][2] = 0; /*  set following energy flow to 0 */
           }
          /*  in every direction there is heat conduction */
          if (   (deriv_o[unit][i+1]<deriv_o[unit][i]) 
              && ( i!=n )
             ) 
           {
           /*  in every direction there is heat conduction */
           /* lambda H20=0.65 W/Km  Teubner Energie k=lambda/d */
           Q_help=0.250*h*( deriv_o[unit][i+1]
                           -deriv_o[unit][i]  );
           Qi[i]  [2] += Q_help;
           Qi[i+1][2] = -Q_help;
           }
          }
        else Qi[i][2]=0; /* if !conv_flag */

        /*----------------------------------------------- ambient losses */
        Qi[i][3]=-kA_node*h*(      deriv_o[unit][i]
                             -out_o[  in_u[unit][9] ]
                                   [ in_nr[unit][9] ]);/*Tamb*/

       } /*end if (first_time_flag) */
      else Qi[i][2]=Qi[i][3]=0;

    } /* end run through nodes-------------------------------------------- */

  /* write  qp_sum[unit][deriv][path] */

  /* ----------------------- energy balance ------------------------------- */
  for (i=1;i<=n;i++)
    {
      energy[i]=init_energy[unit][i];

      if (a) printf("init_energy[unit][1]:%.4f \n"
                    ,init_energy[unit][i]);
      /* write  qp_sum[unit][deriv][path]  Qi = 0 , if no active loop */
      for (i1=1;i1<=8;i1++)    /* qp_sum[unit][node][kind_qp] */
        {
          qp_sum[unit][i][i1]+=Qi[i][i1];

          if (a) printf("Qi[%d][%d]:%.4f \n"
                        ,i
                        ,i1
                        ,Qi[i][i1]
                       );

          energy[i]+=qp_sum[unit][i][i1];
        }
      deriv_n[unit][i]=energy[i]/mc_node;
      if (deriv_n[unit][i]<5) { printf("storage unit:%d temp%d=%.1f "
                                     /*  "Q1:%.1f Q2:%.1f Q3:%.1f Q4:%.1f \n" */
                                       "Q5:%.1f Q6:%.1f Q7:%.1f Q8:%.1f \n" 
                                     ,unit,i,deriv_n[unit][i]
                                     ,Qi[i][5],Qi[i][6],Qi[i][7],Qi[i][8]);
                                     exit (-1);
                                      }
    }

  if (a)
    {
      printf("\n Storage-dump-layer1-------:\n"
             "Q_wt1:   %8.4f \n"
             "Q_convec:%8.4f \n"
             "Q_amb   :%8.4f \n"
             "Q_KW-QWW:%8.4f \n"
             "Q_circ  :%8.4f \n\n",
             Qi[1][1],
             Qi[1][2],
             Qi[1][3],
             Qi[1][4]+Qi[1][5],
             Qi[1][6]+Qi[1][7]
            );
    };

  /*--------------set outputs----------------------*/

  /*wt2 dQ_Koll_out */
  if (call_wt2_flag)
   {
   if ( par[unit][22]!=1 ) /*  mode1: 0 WT-->WT2 existent */
     {
     out_n[unit][1]=pipe_out_n[storage_counter*2+1][1];   /* Q_out pipe wt2 */
     out_n[unit][2]=pipe_out_n[storage_counter*2+1][2];   /* mp_wt2, also set in set
     /* out_n[unit][20]=out_n[un_max+nr-1][12]/h; */
     out_n[unit][20]=(out_n[ in_u [unit][1] ]
                           [ in_nr[unit][1] ]-out_n[unit][1])/h;
     }
   else  /* mode 1: 0 WT */  
     {
      out_n[unit][1]=mp_cp_WT2*h*deriv_o[unit][wt2_out];
      out_n[unit][2]=out_n[ in_u [unit][2] ]
                          [ in_nr[unit][2] ];
      /* P wt2 */
      out_n[unit][20]=(out_n[ in_u [unit][1] ]
                            [ in_nr[unit][1] ]-out_n[unit][1])/h;
      if (0) /* out_n[unit][20]<0) */
       { 
       printf("Qin:%.1f Tin:%.1f QWT2:%.1f Qout:%.1f Tout:%.1f \n"
              ,out_n[ in_u [unit][1] ] [ in_nr[unit][1] ]
              ,out_n[ in_u [unit][1] ] [ in_nr[unit][1] ]/mp_cp_WT2
              ,out_n[unit][20]
              ,out_n[unit][1]
              ,out_n[unit][1]/mp_cp_WT2);
       /*for (i2=1;i2<=10;i2+=1) printf("temp%d:%.1f\n",i2,deriv_n[unit][i2]);
        */
       }
     }                     
   }
  if (KW_flag)
   {
    out_n[unit][3]=-Qi[1][5]; /* Q_WW_out */
    /* mp_WW_out later set in set_mp()*/
    out_n[unit][4]=out_n[  in_u [unit][4] ]       /*mp_KW */
                        [  in_nr[unit][4] ];
    /*P_WW*/
    out_n[unit][22]=(out_n[unit][3]
                    -out_n[  in_u [unit][3] ] 
                          [  in_nr[unit][3] ])/h;
   }


  /*System Energy and Power */
  out_n[unit][6]=0; /* System-Input-Energy [Ws]*/
  /* System-Output-Energy [Ws]  Q_T_Umg */
  if (first_time_flag)
   {
   out_n[unit][7]=0;
   for (i=1;i<=n;i++)
    out_n[unit][7]+=-Qi[i][3]; /* Q_loss amb */
   }

  /* internal energy [Ws]: energy storage + wt1 + wt2  */

  out_n[unit][8]=0; 
  for (i=1;i<=n;i++)  /* sum of each layer */
    out_n[unit][8]+=energy[i];

  if ( (wt1_in) &&  ( par[unit][22]==0 ) )
    {
    out_n[unit][8]+= pipe_out_n[storage_counter*2+0]  [8];
    }

  if ( (wt2_in) && (par[unit][22]!=1) ) /*  mode1: 0 WT-->WT2 existent */
    out_n[unit][8]+= pipe_out_n[storage_counter*2+0][8];             


  if (a)
  printf("out6,7,8 of unit %d  :%.4f  :%.4f  %.4f\n"
         ,unit
         ,out_n[unit][6]
         ,out_n[unit][7]
         ,out_n[unit][8]
        );

  if (call_wt1_flag) 
    if ( par[unit][22]==0 )
     {
     out_n[unit][9] =pipe_out_n[storage_counter*2+0][1];    /*Q_out wt1 */
     out_n[unit][10]=pipe_out_n[storage_counter*2+0][2];   /*mp_wt heater */
     out_n[unit][21]=pipe_out_n[storage_counter*2+0][12]/h; /* P wt1 */
     }
    else  /* mode !=0 */
     {
      out_n[unit][9]=mp_cp_WT1*h*deriv_o[unit][wt1_out];
      out_n[unit][10]=out_n[ in_u [unit][6] ]
                           [ in_nr[unit][6] ];
      /* P wt1 */                          
      out_n[unit][21]=(out_n[ in_u [unit][5] ]
                            [ in_nr[unit][5] ]-out_n[unit][9])/h;
      if (a) 
      {
       printf(" mp_out:%.2f wt1_out:%d \n",out_n[unit][10]*3600,wt1_out);
       printf(" Qi[3..5]sum:%.2f\n",Qi[3][1]+Qi[4][1]+Qi[5][1]);
       printf(" T3..5:%.4f,%.4f,%.4f,%.4f\n"
                ,deriv_o[unit][3],deriv_o[unit][4]
               ,deriv_o[unit][5],deriv_o[unit][6]);
       printf("Q_inWT1:%.2f,Q_out:%.2f \n",out_n[ in_u [unit][5] ]
                                                [ in_nr[unit][5] ],out_n[unit][9]);
       printf("Q_in-Q_out:%.2f \n",out_n[ in_u [unit][5] ]
                                        [ in_nr[unit][5] ]-out_n[unit][9]);
      }
     }

  /* T_WW */
  out_n[unit][11]=deriv_n[unit][ (int) par[unit][17] ]; /* Tsensor1: node 1 T_WW */
  out_n[unit][12]=deriv_n[unit][ (int) par[unit][18] ]; /* Tsensor2 */
  out_n[unit][13]=deriv_n[unit][ (int) par[unit][19] ]; /* Tsensor3 */
  out_n[unit][14]=deriv_n[unit][ (int) par[unit][20] ]; /* Tsensor4 */

  /* T of layer wt1_in  */
  if (call_wt1_flag)
    {
      out_n[unit][15]=0;
      for (i=wt1_in;i<=wt1_out;i+=step_wt1)
        out_n[unit][15]+=deriv_n[unit][i];
      out_n[unit][15]/=(abs (wt1_out-wt1_in)+1 );
    }

  /* T_of layer wt2_in  */
  if (call_wt2_flag)
    {
      out_n[unit][16]=0; /* set mean of T-nodes for T_wt2_amb */
      for (i=wt2_in;i<=wt2_out;i+=step_wt2)
        out_n[unit][16]+=deriv_n[unit][i];
      out_n[unit][16]/=(abs (wt2_out-wt2_in)+1 );
      
      /* UNISOL, internal loop and internal  */
      
      if ( (int) par [unit][16]==0 )  /* take internal FIT with T_wt2_amb   */
        out_n[unit][16]=deriv_n[unit][wt2_out]; /* out_n[ in_u [unit][10] ]
                                                       [ in_nr[unit][10] ];  */  
    }

  if (circ_flag)
    {
     /* Q_circ_out */
     out_n[unit][17]=-Qi[(int) par[unit][23]][7]; /* Q_circ_out=par23 */
     out_n[unit][18]=out_n[ in_u [unit][8] ]
                          [ in_nr[unit][8] ];
     /*Power of circ */                   
     out_n[unit][27]=(-out_n[unit][17]+out_n[ in_u [unit][7] ]
                                            [ in_nr[unit][7] ])/h;
    }
        
  /*first exe time out_n[unit][19] */  
  
  /* [W] Power wt2 (losses of pipe)
     out_n[unit][20] earlier set in out 4  */ 
  
  /* [W] Power of heating in he1
    out_n[unit][21] earlier set at out 9,10 */
  
  /* [W] Power of load_ww */
  /* earlier set in out 3 
    out_n[unit][22]= */
  
  /* [W] Power of losses to ambient */
  if (first_time_flag) { out_n[unit][23]=0;
                         for (i=1;i<=n;i++) out_n[unit][23]-=Qi[i][3]; /* [Ws] */
                         out_n[unit][23]/=h; /* use power [W] */
                       }         
  
  /* [C] temp wt2 out */                    
  out_n[unit][24]=pipe_out_n[storage_counter*2+1][3];         /* T_WT2 */

  out_n[unit][25]=0;        /* Tmean */
  for (i=1;i<=n;i++) out_n[unit][25]+=deriv_n[unit][i];
  out_n[unit][25]/=n;
  
  out_n[unit][26]=pipe_out_n[storage_counter*2+1][8]; /* Q_lower part */
  for (i=6;i<=n;i++) out_n[unit][26]+=energy[i];

  /*out27: [W] Power of circ,  earlier set at circ flag */
  out_n[unit][28]=pipe_out_n[storage_counter*2+1][11]; /* temp wt2 inlet */
  out_n[unit][29]=pipe_out_n[storage_counter*2+0][11]; /* temp wt1 inlet */
  out_n[unit][30]=pipe_out_n[storage_counter*2+0][ 3]; /* temp wt1 outlet*/

  /* node temperature */
  for (i=1;i<=9;i++) out_n[unit][30+i]=deriv_n[unit][i];
  
  return 1;
}
